Package: xreg2
Type: Package
Title: Flexible Maximum Likelihood Regression with Gradient-Based
        Optimisation
Version: 0.1.4
Authors@R: person("Kim", "Rand", email = "krand@krand.no", role = c("aut", "cre"))
Description: Fits flexible maximum likelihood regression models supporting censored,
    interval, and hybrid continuous/dichotomous data. Provides explicit analytic and
    numerical gradient computation, random intercept models via Gauss-Hermite
    quadrature, and multiple distribution families.
URL: https://github.com/MathsinHealth/xreg2
Depends: R (>= 4.0.0)
Imports: ucminf, numDeriv
Suggests: optimx, testthat (>= 3.0.0)
License: GPL-2
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-23 12:56:33 UTC; Eier
Author: Kim Rand [aut, cre]
Maintainer: Kim Rand <krand@krand.no>
Repository: CRAN
Date/Publication: 2026-04-28 18:10:02 UTC
Built: R 4.6.0; ; 2026-04-28 20:30:22 UTC; unix

Package: xreg2
Type: Package
Title: Flexible Maximum Likelihood Regression with Gradient-Based
        Optimisation
Version: 0.1.4
Authors@R: person("Kim", "Rand", email = "krand@krand.no", role = c("aut", "cre"))
Description: Fits flexible maximum likelihood regression models supporting censored,
    interval, and hybrid continuous/dichotomous data. Provides explicit analytic and
    numerical gradient computation, random intercept models via Gauss-Hermite
    quadrature, and multiple distribution families.
URL: https://github.com/MathsinHealth/xreg2
Depends: R (>= 4.0.0)
Imports: ucminf, numDeriv
Suggests: optimx, testthat (>= 3.0.0)
License: GPL-2
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-23 12:56:33 UTC; Eier
Author: Kim Rand [aut, cre]
Maintainer: Kim Rand <krand@krand.no>
Repository: CRAN
Date/Publication: 2026-04-28 18:10:02 UTC
Built: R 4.6.0; ; 2026-05-01 20:32:51 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(AIC,xreg2)
S3method(ARE,data.frame)
S3method(ARE,matrix)
S3method(ARE,numeric)
S3method(ARE,two_numeric)
S3method(BIC,xreg2)
S3method(MAE,data.frame)
S3method(MAE,matrix)
S3method(MAE,numeric)
S3method(MAE,two_numeric)
S3method(MARE,data.frame)
S3method(MARE,matrix)
S3method(MARE,numeric)
S3method(MARE,two_numeric)
S3method(MBE,data.frame)
S3method(MBE,matrix)
S3method(MBE,numeric)
S3method(MBE,two_numeric)
S3method(RMSE,data.frame)
S3method(RMSE,matrix)
S3method(RMSE,numeric)
S3method(RMSE,two_numeric)
S3method(RMSRE,data.frame)
S3method(RMSRE,matrix)
S3method(RMSRE,numeric)
S3method(RMSRE,two_numeric)
S3method(ReMSE,data.frame)
S3method(ReMSE,matrix)
S3method(ReMSE,numeric)
S3method(ReMSE,two_numeric)
S3method(U95,data.frame)
S3method(U95,matrix)
S3method(U95,numeric)
S3method(U95,two_numeric)
S3method(appendmodel,formula)
S3method(c,xreg2Control)
S3method(c,xreg2ControlList)
S3method(change_target,formula)
S3method(erMAX,data.frame)
S3method(erMAX,matrix)
S3method(erMAX,numeric)
S3method(erMAX,two_numeric)
S3method(fisher_icc,data.frame)
S3method(fisher_icc,matrix)
S3method(fisher_icc,numeric)
S3method(fisher_icc,two_numeric)
S3method(just_CCC,data.frame)
S3method(just_CCC,matrix)
S3method(just_CCC,numeric)
S3method(just_CCC,two_numeric)
S3method(predict,xreg2)
S3method(print,xreg2)
S3method(summary,xreg2)
S3method(tstat,data.frame)
S3method(tstat,matrix)
S3method(tstat,numeric)
S3method(tstat,two_numeric)
export(ARE)
export(ARE_conform)
export(LCCC)
export(MAE)
export(MAPE)
export(MARE)
export(MBE)
export(RMSE)
export(RMSRE)
export(ReMSE)
export(ReMSE_conform)
export(U95)
export(combmean)
export(combmeans)
export(combmssds)
export(combsd)
export(combsds)
export(comp_table)
export(erMAX)
export(fisher_icc)
export(fixdist)
export(getMean)
export(getSD)
export(getVar)
export(getd)
export(getf)
export(getp)
export(getq)
export(getr)
export(is.wholenumber)
export(just_CCC)
export(ordD)
export(ordE)
export(ordEp)
export(pRMSE)
export(quant_fun)
export(quant_fun_groups)
export(quant_funs)
export(quant_funs_groups)
export(quant_funst)
export(rRMSE)
export(t_test2)
export(tstat)
export(xreg2_cont)
export(xreg2_cont_normal)
export(xreg2_cont_r_lnorm)
export(xreg2_cont_r_normal)
export(xreg2_control)
export(xreg2_dich_logistic)
export(xreg2_fit)
export(xreg2_fit_s)
export(xreg2_hyreg)
export(xreg2_optim)
importFrom(stats,AIC)
importFrom(stats,BIC)
importFrom(stats,aggregate)
importFrom(stats,as.formula)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,dbeta)
importFrom(stats,dexp)
importFrom(stats,dgamma)
importFrom(stats,dlnorm)
importFrom(stats,dnorm)
importFrom(stats,dweibull)
importFrom(stats,formula)
importFrom(stats,integrate)
importFrom(stats,na.omit)
importFrom(stats,optim)
importFrom(stats,pbeta)
importFrom(stats,pexp)
importFrom(stats,pgamma)
importFrom(stats,plnorm)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,pweibull)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(utils,modifyList)

#' Fit a flexible maximum-likelihood regression model
#'
#' @title Fit xreg2 Model
#' @description Fits a flexible maximum-likelihood regression model supporting
#'   censored, interval, and hybrid data. Accepts a list of
#'   \code{\link{xreg2_control}} objects and a matching named list of
#'   data frames.
#'
#' @param controlList An \code{xreg2ControlList} (produced by combining
#'   \code{xreg2_control} objects with \code{c()}), an \code{xreg2_obj} (for
#'   re-optimisation), or an \code{xreg2}/\code{xreg} fitted object (for
#'   re-fitting).
#' @param dataList Named list of data frames, one per component. Names must
#'   match those of \code{controlList}.
#' @param start_values Optional named numeric vector of starting values,
#'   overriding those in \code{controlList}.
#' @param fixed_values Named numeric vector of parameter values held fixed.
#' @param latent_classes Integer; number of latent classes. \code{0} disables.
#' @param latent_class_parameters Character vector of parameter names that
#'   vary by latent class.
#' @param latent_id_colname Character name of the grouping column for latent
#'   classes.
#' @param return_type Character; \code{"fit"} (default), \code{"control"},
#'   \code{"df"}, or \code{"predict"}.
#' @param print_sum Logical; print summary during optimisation. Default
#'   \code{FALSE}.
#' @param method Character; optimisation method. Default \code{"ucminf"}.
#' @param hessian Logical; compute Hessian for standard errors. Default
#'   \code{TRUE}.
#' @param run_from Internal; character string identifying the calling wrapper
#'   (e.g. \code{"hyreg"}). Used to suppress certain warnings when called
#'   from higher-level functions.
#' @param ... Additional arguments passed to the optimiser.
#'
#' @return An object of class \code{"xreg2"} with fitted coefficients,
#'   standard errors, log-likelihoods, and counts.
#'
#' @seealso \code{\link{xreg2_control}}, \code{\link{xreg2_optim}},
#'   \code{\link{predict.xreg2}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
#' ctrl <- xreg2_control(
#'   formulas     = list(y ~ INTERCEPT + x * SLOPE),
#'   start_values = c(INTERCEPT = 0, SLOPE = 0),
#'   name         = "main"
#' )
#' fit  <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))
#' coef(fit)
#' }
#'
#' @export
xreg2_fit <- function(controlList,
                       dataList                = NULL,
                       start_values            = numeric(),
                       fixed_values            = numeric(),
                       latent_classes          = 0,
                       latent_class_parameters = character(),
                       latent_id_colname       = character(),
                       return_type             = "fit",
                       print_sum               = FALSE,
                       method                  = "ucminf",
                       hessian                 = TRUE,
                       run_from                = NULL,
                       ...) {
  xreg2_time <- Sys.time()

  if ("xreg2_obj" %in% class(controlList))
    return(xreg2_optim(xreg2obj = controlList, method = method,
                        hessian = hessian, fixed_values = fixed_values, ...))

  if (is.null(dataList)) stop('Argument "dataList" is missing, with no default.')

  dot_args <- list(...)
  dot_args[c("p_fun", "p_aggrgation_fun")] <- NULL

  if (latent_classes < 2) latent_class_parameters <- character()

  if (!is.character(latent_class_parameters)) {
    warning("Latent_class_parameters not character. Latent classes will be disregarded.")
    latent_class_parameters <- character()
    latent_classes          <- 0
    latent_id_colname       <- character()
  }
  if (!is.character(latent_id_colname)) {
    warning("latent_id_colname not character. Latent classes will be disregarded.")
    latent_class_parameters <- character()
    latent_classes          <- 0
    latent_id_colname       <- character()
  }
  if (length(latent_id_colname)) latent_id_colname <- latent_id_colname[1]

  startValues <- attr(controlList, "startValues")

  if (class(controlList)[1] %in% c("xreg", "xreg2")) {
    if (is.na(start_values[1])) start_values <- controlList$coef
    controlList <- controlList$controlList
  }

  formulas    <- controlList
  controlList <- do.call(xreg2_control,
                         match_formals(fun = xreg2_control, formulas = controlList))

  if ("xreg"    %in% class(start_values)) startValues <- start_values$coef
  if ("numeric" %in% class(start_values) && length(start_values) > 0)
    startValues <- start_values

  if ("data.frame" %in% class(dataList)) {
    dataList <- list(dataList)
    names(dataList)[1] <- names(controlList)[1]
    warning(paste0(
      "Dataframe provided where list expected in dataList argument. ",
      "The dataframe has been enclosed in a list, and named \"",
      names(controlList)[1], "\" to match the first xreg2Control-object."))
  }

  if (!all(names(dataList) %in% names(controlList)))
    stop(paste("No control list provided for ",
               paste(names(dataList)[!names(dataList) %in% names(controlList)],
                     collapse = ", ")))

  if (is.null(names(dataList))) {
    if (length(dataList) <= length(controlList)) {
      warning("Provided datalist was unnamed. Names coerced from controlList")
      names(dataList) <- names(controlList)[seq_len(length(dataList))]
    } else {
      stop("Provided datalist was unnamed, and longer than provided controllist. Exiting.")
    }
  }

  if (length(unique(names(startValues))) < length(startValues)) {
    warning(paste0(
      "More than one start value assigned for parameters (",
      paste(names(startValues)[-match(unique(names(startValues)), names(startValues))],
            collapse = ", "), "). First value used."))
    startValues <- startValues[unique(names(startValues))]
  }

  if (length(fixed_values)) {
    if (class(fixed_values)[1] %in% c("xreg", "xreg2")) {
      fixed_values <- fixed_values$coef
    }
    if ("data.frame" %in% class(fixed_values)) {
      fixed_df     <- fixed_values[, 1:2]
      fixed_values <- fixed_values[, 1]
      names(fixed_values) <- rownames(fixed_df)
    } else {
      fixed_df <- data.frame(Estimate    = fixed_values,
                             `Std. Error` = rep(NA, NROW(fixed_values)))
      colnames(fixed_df) <- c("Estimate", "Std. Error")
      rownames(fixed_df) <- names(fixed_values)
    }
  } else {
    fixed_df <- data.frame(a = numeric(), b = numeric())
  }
  colnames(fixed_df) <- c("Estimate", "Std. Error")
  fixedValues <- numeric()

  aggr            <- TRUE
  return_lik_df   <- (return_type == "df")
  return_first    <- (return_type == "first")
  if (return_type == "predict") {
    aggr          <- FALSE
    return_lik_df <- TRUE
    return_type   <- "first_df"
  }
  if (return_type == "first_df") {
    return_first    <- TRUE
    return_first_df <- TRUE
  } else {
    return_first_df <- FALSE
  }
  if (return_type == "precalc_df") {
    return_first    <- TRUE
    return_lik_df   <- FALSE
    return_first_df <- TRUE
  }

  prep_out        <- list()
  newDataList     <- list()
  all_defined_vars <- character()
  undefined_vars  <- character()
  valuevars       <- list()

  for (dataName in names(dataList)) {
    rel_col      <- vector()
    thisControl  <- controlList[[dataName]]
    data_df      <- dataList[[dataName]]
    orig_cols    <- colnames(data_df)
    valueVar     <- NA
    defined_vars <- colnames(data_df)
    local_vars   <- character()

    for (fnum in seq_along(thisControl$formulas)) {
      formula    <- thisControl$formulas[[fnum]]
      targetName <- as.character(formula[[2]])
      rel_col    <- c(rel_col, all.vars(formula)[!all.vars(formula) %in% rel_col])

      if (is.na(valueVar)) {
        if (targetName %in% colnames(data_df) ||
            length(grep(paste0(targetName, "\\."), colnames(data_df), value = TRUE)) == 2) {
          valuevars[[dataName]] <- targetName
          valueVar              <- targetName
        }
      }
      rhs_vars       <- all.vars(formula[[3]])
      undefined_vars <- c(undefined_vars, rhs_vars[!rhs_vars %in% defined_vars])
      all_defined_vars <- c(all_defined_vars,
                            undefined_vars[!undefined_vars %in% all_defined_vars])
      if (!targetName %in% c(colnames(data_df), valueVar))
        defined_vars <- c(defined_vars, targetName)
      local_vars <- c(local_vars, all.vars(formula), targetName)
    }

    local_vars <- local_vars[!local_vars %in% colnames(data_df)]
    controlList[[dataName]][["defined_vars"]] <- local_vars

    censor_bounds <- thisControl$censor_bounds
    grep_vals     <- grep(paste0(valueVar, "\\."), colnames(data_df), value = TRUE)

    if (length(grep_vals) > 2)
      stop("More than two columns matching target variable.")

    if (length(grep_vals) == 2) {
      if (all((data_df[, grep_vals[1]] - data_df[, grep_vals[2]]) >= 0)) {
        upper_bound_var <- grep_vals[1]
        lower_bound_var <- grep_vals[2]
      } else if (all((data_df[, grep_vals[1]] - data_df[, grep_vals[2]]) <= 0)) {
        upper_bound_var <- grep_vals[2]
        lower_bound_var <- grep_vals[1]
      } else {
        stop(paste0("Two columns matched the left-hand side: ",
                    grep_vals[1], " and ", grep_vals[2],
                    ". However, neither was always >= the other, so they cannot ",
                    "be used as upper/lower bounds. Quitting."))
      }
    } else {
      if (!valueVar %in% colnames(data_df)) {
        stop("No value columns specified.")
      } else {
        upper_bound_var <- lower_bound_var <- valueVar
      }
    }

    controlList[[dataName]]$upper_bound_var <- upper_bound_var
    controlList[[dataName]]$lower_bound_var <- lower_bound_var
    controlList[[dataName]]$valueVar        <- valueVar

    data_df[, c("internal_ub", "internal_lb")] <- data_df[, c(upper_bound_var, lower_bound_var)]
    formula[[2]] <- substitute(Xb)
    e_params     <- list(...)

    if (length(latent_id_colname)) {
      if (latent_id_colname %in% colnames(data_df))
        data_df$internal_classgroup <- data_df[, latent_id_colname]
    } else {
      data_df$internal_classgroup <- paste0(dataName, ".", seq_len(NROW(data_df)))
    }

    if ("id" %in% colnames(data_df)) data_df$internal_id <- as.integer(data_df$id)

    data_df <- within(data_df, {
      both_na    <- is.na(internal_ub) * is.na(internal_lb)
      internal_ub[internal_ub >= max(thisControl$censor_bounds)] <- Inf
      internal_lb[internal_lb <= min(thisControl$censor_bounds)] <- -Inf
      internal_lb[internal_lb >= max(thisControl$censor_bounds)] <- max(thisControl$censor_bounds)
      internal_ub[internal_ub <= min(thisControl$censor_bounds)] <- min(thisControl$censor_bounds)
      internal_ub[is.na(internal_ub)] <- Inf
      internal_lb[is.na(internal_lb)] <- -Inf
      internal_type              <- rep(3, NROW(internal_ub))
      internal_type[internal_ub == internal_lb]  <- 1
      internal_type[internal_ub == Inf]          <- 0
      internal_type[internal_lb == -Inf]         <- 2
      internal_count <- 1
    })

    if (sum(data_df$both_na)) {
      warning(paste0(sum(data_df$both_na), " rows in which ", valueVar,
                     " was NA were removed."))
      data_df <- data_df[!data_df$both_na, -(NCOL(data_df))]
    }

    if (!is.na((weights_var <- thisControl$weights_var)) &&
        weights_var %in% colnames(data_df))
      data_df$internal_count <- data_df[, weights_var]

    relevant_columns <- c(rel_col,
                          colnames(data_df)[grep("internal_", colnames(data_df))])

    if (aggr) {
      new_df             <- data_df[, colnames(data_df) %in% relevant_columns]
      new_df$internal_unique <- as.numeric(as.factor(
        apply(new_df, MARGIN = 1, FUN = "paste", collapse = "")))
      unique_df              <- unique(new_df)
      unique_df$internal_count <- aggregate(
        new_df$internal_count,
        by  = list(new_df$internal_unique),
        FUN = "sum")[unique_df$internal_unique, 2]
      newDataList[[dataName]] <- unique_df
    } else {
      newDataList[[dataName]] <- data_df
    }

    controlList[[dataName]]$orig_cols <- orig_cols
    startValues <- c(startValues,
                     thisControl$start_values[!names(thisControl$start_values) %in% names(startValues)])
    fixedValues <- c(fixedValues,
                     thisControl$fixed_values[!names(thisControl$fixed_values) %in% names(fixedValues)])
    all_defined_vars <- c(all_defined_vars,
                          defined_vars[!defined_vars %in% all_defined_vars])
    controlList[[dataName]]$obs_types <- c(
      Uncensored    = sum(data_df$internal_count * (data_df$internal_ub == data_df$internal_lb)),
      `Left censored`  = sum(data_df$internal_count * (data_df$internal_ub > data_df$internal_lb) *
                               (data_df$internal_ub < Inf) * (data_df$internal_lb == -Inf)),
      `Right censored` = sum(data_df$internal_count * (data_df$internal_ub > data_df$internal_lb) *
                               (data_df$internal_lb > -Inf) * (data_df$internal_ub == Inf)),
      Intervals     = sum(data_df$internal_count * (data_df$internal_ub > data_df$internal_lb) *
                            (data_df$internal_lb > -Inf) * (data_df$internal_ub < Inf)))
  }

  if (length(fixedValues)) {
    fixed_values[names(fixedValues)] <- fixedValues
    tmp <- rep(NA, NROW(fixed_values))
    names(tmp) <- names(fixed_values)
    tmp[rownames(fixed_df)] <- fixed_df[, 2]
    fixed_df <- data.frame(Estimate    = fixed_values,
                           `Std. Error` = tmp)
    colnames(fixed_df) <- c("Estimate", "Std. Error")
  }

  startValues[names(fixed_values)[names(fixed_values) %in% names(startValues)]] <-
    fixed_values[names(fixed_values) %in% names(startValues)]

  undefined_vars <- unique(undefined_vars[!undefined_vars %in%
                                            c(names(startValues), latent_class_parameters)])
  undefined_vars <- undefined_vars[!undefined_vars %in% c("Xb", names(fixedValues))]

  if (length(undefined_vars)) {
    tmp        <- paste(undefined_vars, collapse = ", ")
    used_from  <- if (!is.null(run_from)) run_from else "xreg2"
    if (!used_from == "hyreg")
      warning(paste0(
        "Used formulas contained variables not found in start_values or column names (",
        tmp, "). These will be initiated with a starting value of 0.1. ",
        "Defining start values either in xreg2_control or in the call to xreg2_fit ",
        "is highly recommended."))
    tmp <- rep(0.1, length(undefined_vars))
    names(tmp) <- undefined_vars
    startValues <- c(startValues, tmp)
  }

  if (latent_classes > 1) {
    for (lc in latent_class_parameters) {
      tmp <- 0.1
      names(tmp) <- lc
      if (lc %in% names(startValues)) {
        tmp         <- startValues[lc]
        startValues <- startValues[-which(names(startValues) == lc)]
        all_defined_vars <- all_defined_vars[-which(all_defined_vars == lc)]
      }
      for (i in seq_len(latent_classes)) {
        lci <- paste(lc, i, sep = ".")
        if (!lci %in% names(startValues)) {
          startValues <- c(startValues, tmp)
          names(startValues)[length(startValues)] <- lci
        }
        if (!lci %in% all_defined_vars) all_defined_vars <- c(all_defined_vars, lci)
      }
    }
  }

  if (length((tmp <- names(startValues[!names(startValues) %in% all_defined_vars])))) {
    startValues <- startValues[!names(startValues) %in% tmp]
    tmp <- paste(tmp, collapse = ", ")
    warning(paste0("Start values defined for variables (", tmp,
                   ") not found in column names or formulas. These have been removed."))
  }

  user_ctrl <- if ("optim_control" %in% names(e_params)) {
    dot_args["optim_control"] <- NULL
    e_params[["optim_control"]]
  } else list()

  default_ctrl <- switch(method,
    "L-BFGS-B"    = list(maxit = 10000, factr  = 1e-11,  lmm    = 10),
    "BFGS"         = list(maxit = 10000, reltol = 1e-11,  abstol = 1e-9),
    "CG"           = list(maxit = 10000, reltol = 1e-11),
    "Nelder-Mead"  = list(maxit = 10000, reltol = 1e-11),
    list()
  )
  ctrl <- modifyList(default_ctrl, user_ctrl)

  startValues[names(fixed_values)] <- fixed_values

  dtns <- names(newDataList)
  names(dtns) <- dtns

  greek <- c("ALPHA", "BETA", "GAMMA", "DELTA", "EPSILON", "ZETA", "ETA",
             "THETA", "IOTA", "KAPPA", "LAMBDA", "MU", "NU", "XI", "OMIKRON",
             "PI", "RHO", "SIGMA", "TAU", "UPSILON", "PHI", "CHI", "PSI", "OMEGA")

  xreg2lobj <- lapply(X = dtns, FUN = function(dtn) {
    thisfull <- newDataList[[dtn]]
    thisfull <- thisfull[, !colnames(thisfull) %in% "internal_classgroup"]
    tmpfull  <- thisfull[, !colnames(thisfull) %in% c("internal_id", "internal_count",
                                                        "internal_unique")]
    thisunique        <- unique(tmpfull)
    thisfull$umatch   <- multimatch(tmpfull, thisunique)

    sv <- startValues[!names(startValues) %in% names(fixed_values)]

    spvns <- unique(unlist(lapply(c(greek, "INTERCEPT", "SLOPE"),
                                  function(x) grep(x, names(sv), value = TRUE))))
    spvs  <- sv[spvns]
    sv    <- sv[!names(sv) %in% spvns]
    sv    <- c(sv, spvs)

    list(full          = thisfull,
         unique        = thisunique,
         coeff_names   = names(sv),
         formulas      = controlList[[dtn]]$formulas,
         lower         = controlList[[dtn]]$lower,
         upper         = controlList[[dtn]]$upper,
         random_formula= NULL,
         fixed         = fixed_values,
         ll_fun        = controlList[[dtn]]$p_fun,
         grad_fun      = controlList[[dtn]]$grad_fun,
         valuevar      = valuevars[[dtn]],
         startValues   = sv,
         ctrl          = ctrl)
  })
  class(xreg2lobj) <- c("xreg2_obj", "list")

  if (return_type == "control") {
    return(xreg2lobj)
  }

  oout <- do.call(xreg2_optim,
                  c(list(xreg2obj = xreg2lobj, method = method, hessian = hessian),
                    dot_args))
  oout$xreg2_time <- Sys.time() - xreg2_time
  return(oout)
}

# ---------------------------------------------------------------------------

#' Fit an xreg2 model from pre-structured data
#'
#' @title Fit xreg2 Model from Pre-Structured Data
#' @description Simplified version of \code{\link{xreg2_fit}} that accepts
#'   data already containing \code{ub} (upper bound) and \code{lb} (lower
#'   bound) columns instead of requiring data pre-processing.
#'
#' @inheritParams xreg2_fit
#'
#' @return An object of class \code{"xreg2"}.
#'
#' @seealso \code{\link{xreg2_fit}}, \code{\link{xreg2_control}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' # xreg2_fit_s expects pre-structured data with lb and ub columns.
#' # Use lb == ub for exact (uncensored) observations.
#' y    <- rnorm(100, 2, 0.5)
#' df   <- data.frame(y = y, x = rnorm(100), lb = y, ub = y)
#' ctrl <- xreg2_control(
#'   formulas     = list(y ~ INTERCEPT + x * SLOPE),
#'   start_values = c(INTERCEPT = 0, SLOPE = 0, LN_SIGMA = 0),
#'   name         = "main"
#' )
#' fit  <- xreg2_fit_s(controlList = c(ctrl), dataList = list(main = df))
#' }
#'
#' @export
xreg2_fit_s <- function(controlList,
                          dataList                = NULL,
                          start_values            = numeric(),
                          fixed_values            = numeric(),
                          latent_classes          = 0,
                          latent_class_parameters = character(),
                          latent_id_colname       = character(),
                          return_type             = "fit",
                          print_sum               = FALSE,
                          method                  = "ucminf",
                          hessian                 = TRUE,
                          ...) {
  xreg2_time <- Sys.time()

  if ("xreg2_obj" %in% class(controlList))
    return(xreg2_optim(xreg2obj = controlList, method = method,
                        hessian = hessian, fixed_values = fixed_values, ...))

  ctrl <- switch(method,
    "BFGS"    = list(maxit = 10000, abstol = 1e-9,   reltol = 1e-11),
    "L-BFGS-B"= list(maxit = 10000, factr  = 1e-11,  lmm    = 10),
    NULL
  )

  dtns        <- names(dataList)
  names(dtns) <- dtns
  startValues <- start_values

  for (dtn in dtns) {
    startValues <- c(startValues,
                     controlList[[dtn]]$start_values[!names(controlList[[dtn]]$start_values) %in%
                                                       names(startValues)])
  }
  startValues[names(fixed_values)] <- fixed_values

  greek <- c("ALPHA", "BETA", "GAMMA", "DELTA", "EPSILON", "ZETA", "ETA",
             "THETA", "IOTA", "KAPPA", "LAMBDA", "MU", "NU", "XI", "OMIKRON",
             "PI", "RHO", "SIGMA", "TAU", "UPSILON", "PHI", "CHI", "PSI", "OMEGA")

  xreg2lobj <- lapply(X = dtns, FUN = function(dtn) {
    thisfull <- dataList[[dtn]]

    if ("id" %in% colnames(thisfull) && !"internal_id" %in% colnames(thisfull))
      thisfull$internal_id <- thisfull$id

    thisfull$internal_lb <- thisfull$lb
    thisfull$internal_ub <- thisfull$ub
    thisfull$internal_type                                    <- 3
    thisfull$internal_type[thisfull$internal_ub == thisfull$internal_lb] <- 1
    thisfull$internal_type[thisfull$internal_ub == Inf]       <- 0
    thisfull$internal_type[thisfull$internal_lb == -Inf]      <- 2
    thisfull$internal_count <- 1

    if (!is.null(controlList[[dtn]]$weights_var))
      if (!is.na(controlList[[dtn]]$weights_var))
        thisfull$internal_count <- thisfull[[controlList[[dtn]]$weights_var]]

    thisfull <- thisfull[, !colnames(thisfull) %in% "internal_classgroup"]
    tmpfull  <- thisfull[, !colnames(thisfull) %in% c("internal_count", "internal_unique")]
    thisunique        <- unique(tmpfull)
    thisfull$umatch   <- multimatch(tmpfull, thisunique)

    sv <- startValues[!names(startValues) %in% names(fixed_values)]

    spvns <- unique(unlist(lapply(c(greek, "INTERCEPT", "SLOPE"),
                                  function(x) grep(x, names(sv), value = TRUE))))
    spvs  <- sv[spvns]
    sv    <- sv[!names(sv) %in% spvns]
    sv    <- c(sv, spvs)

    fv <- if (!length(fixed_values)) NULL else fixed_values

    list(full          = thisfull,
         unique        = thisunique,
         coeff_names   = names(sv),
         formulas      = controlList[[dtn]]$formulas,
         lower         = controlList[[dtn]]$lower,
         upper         = controlList[[dtn]]$upper,
         random_formula= NULL,
         fixed         = fv,
         ll_fun        = controlList[[dtn]]$p_fun,
         grad_fun      = controlList[[dtn]]$grad_fun,
         valuevar      = NULL,
         startValues   = sv,
         ctrl          = ctrl)
  })
  class(xreg2lobj) <- c("xreg2_obj", "list")

  if (return_type == "control") {
    return(xreg2lobj)
  }

  oout <- xreg2_optim(xreg2lobj, method = method, hessian = hessian, ...)
  oout$xreg2_time <- Sys.time() - xreg2_time
  return(oout)
}

# ---------------------------------------------------------------------------

#' Optimise an xreg2 model object
#'
#' @title Optimise an xreg2 Model
#' @description Low-level optimisation routine called by \code{\link{xreg2_fit}}
#'   and \code{\link{xreg2_fit_s}}. Can also be called directly on an
#'   \code{xreg2_obj} to re-optimise with different settings.
#'
#'   Optimizer routing:
#'   \itemize{
#'     \item \code{"ucminf"} (default) — uses \pkg{ucminf} with central-difference
#'       gradients and high-precision tolerances.
#'     \item \code{"L-BFGS-B"}, \code{"BFGS"}, \code{"CG"}, \code{"Nelder-Mead"},
#'       \code{"SANN"}, \code{"Brent"} — uses \code{stats::optim()}.
#'       Box constraints (\code{lower}/\code{upper} set in
#'       \code{\link{xreg2_control}}) are enforced for \code{"L-BFGS-B"}.
#'     \item Any other method string — requires the \pkg{optimx} package (listed
#'       in \code{Suggests}) and dispatches via \code{optimx::optimr()}.
#'   }
#'
#'   Method-specific control defaults (high-precision) are set automatically and
#'   can be overridden by passing \code{optim_control = list(...)} to
#'   \code{\link{xreg2_fit}}.
#'
#' @param xreg2obj An \code{xreg2_obj} list produced by
#'   \code{\link{xreg2_fit}} with \code{return_type = "control"}.
#' @param method Character; optimisation method. \code{NULL} defaults to
#'   \code{"ucminf"}.
#' @param hessian Logical; compute Hessian. Default \code{TRUE}.
#' @param fixed_values Optional named numeric vector of additional fixed values.
#' @param optim_control Optional named list of control parameters that override
#'   the method-specific defaults set by \code{\link{xreg2_fit}}.  For example,
#'   \code{optim_control = list(maxit = 500)} tightens the iteration budget.
#'   Passed directly to \code{\link[ucminf]{ucminf}}, \code{\link[stats]{optim}},
#'   or \code{optimx::optimr} depending on \code{method}.
#' @param ... Additional arguments passed to the optimiser objective function.
#'
#' @return An object of class \code{"xreg2"}.
#'
#' @importFrom stats optim
#' @importFrom utils modifyList
#' @seealso \code{\link{xreg2_fit}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
#' ctrl <- xreg2_control(
#'   formulas     = list(y ~ INTERCEPT + x * SLOPE),
#'   start_values = c(INTERCEPT = 0, SLOPE = 0),
#'   name         = "main"
#' )
#' obj          <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
#'                           return_type = "control")
#' fit          <- xreg2_optim(obj, method = "ucminf")
#' fit_lbfgsb   <- xreg2_optim(obj, method = "L-BFGS-B")
#' }
#'
#' @export
xreg2_optim <- function(xreg2obj, method = NULL, hessian = TRUE,
                          fixed_values = NULL, optim_control = NULL, ...) {

  if (is.null(method)) method <- "ucminf"

  pars <- xreg2obj[[1]]$startValues

  if (!is.null(fixed_values)) {
    fixedvars <- xreg2obj[[1]]$fixed
    fixedvars[names(fixed_values)] <- fixed_values
    xreg2obj[[1]]$fixed      <- fixedvars
    pars <- pars[!names(pars) %in% names(fixedvars)]
    xreg2obj[[1]]$startValues <- pars
  }

  # Methods supported natively by stats::optim()
  OPTIM_METHODS <- c("nelder-mead", "bfgs", "cg", "l-bfgs-b", "sann", "brent")

  # Base control from xreg2_fit (method-specific defaults already merged with
  # any optim_control the user passed to xreg2_fit).  A further optim_control
  # passed directly to xreg2_optim() is applied on top.
  base_ctrl <- if (!is.null(xreg2obj[[1]]$ctrl)) xreg2obj[[1]]$ctrl else list()
  if (!is.null(optim_control))
    base_ctrl <- modifyList(base_ctrl, optim_control)

  optim_time_start <- Sys.time()

  if (length(pars)) {
    if (tolower(method) == "ucminf") {
      # ucminf-specific high-precision defaults; user ctrl overrides
      ucminf_defaults <- list(grad     = "central",
                              gradstep = c(1e-7, 1e-8),
                              maxeval  = 10000,
                              grtol    = 1e-8)
      ctrl <- modifyList(ucminf_defaults, base_ctrl)
      oout <- ucminf::ucminf(par = pars, fn = negll_list2, gr = gradll_list2,
                             dtalist = xreg2obj, hessian = hessian,
                             control = ctrl, ...)

    } else if (tolower(method) == "l-bfgs-b") {
      upper <- rep(Inf,  length(pars))
      lower <- rep(-Inf, length(pars))
      names(upper) <- names(lower) <- names(pars)

      uppers <- xreg2obj[[1]]$upper
      lowers <- xreg2obj[[1]]$lower
      uppers <- uppers[which(names(uppers) %in% names(pars))]
      lowers <- lowers[which(names(lowers) %in% names(pars))]
      upper[names(uppers)] <- uppers
      lower[names(lowers)] <- lowers

      oout <- stats::optim(par = pars, fn = negll_list2, gr = gradll_list2,
                           dtalist = xreg2obj,
                           method = "L-BFGS-B",
                           lower = lower, upper = upper,
                           control = base_ctrl, hessian = hessian, ...)

    } else if (tolower(method) %in% OPTIM_METHODS) {
      oout <- stats::optim(par = pars, fn = negll_list2, gr = gradll_list2,
                           dtalist = xreg2obj,
                           method = method,
                           control = base_ctrl, hessian = hessian, ...)

    } else {
      if (!requireNamespace("optimx", quietly = TRUE))
        stop(sprintf(
          paste0('Optimizer method "%s" is not supported by stats::optim().\n',
                 'Install the "optimx" package to enable additional methods:\n',
                 '  install.packages("optimx")'),
          method))
      oout <- optimx::optimr(par = pars, fn = negll_list2, gr = gradll_list2,
                             method = method, dtalist = xreg2obj,
                             hessian = hessian, control = base_ctrl, ...)
    }
    oout$fixed <- xreg2obj[[1]]$fixed
  } else {
    oout <- list(par   = c(pars, fixedvars),
                 fixed = fixedvars,
                 value = negll_list2(par = c(pars, fixedvars), dtalist = xreg2obj))
    hessian <- FALSE
  }

  oout$SE <- rep(NA, length(oout$par))
  names(oout$SE) <- names(oout$par) <- xreg2obj[[1]]$coeff_names

  if (hessian) {
    oout$vcov <- solve(oout$hessian)
    oout$SE[] <- sqrt(diag(oout$vcov))
  }

  oout$formulas <- lapply(xreg2obj, function(x) x$formulas)

  coef     <- oout$par
  oout$parf <- c(coef, oout$fixed)
  ncoef    <- length(coef)
  nfixed   <- length(oout$fixed)

  dtns        <- names(xreg2obj)
  names(dtns) <- dtns

  oout$minima <- sapply(dtns, function(dtn) negll_list2(oout$par, xreg2obj[dtn], ...))
  oout$minima <- c(oout$minima, total = sum(oout$minima))
  oout$counts <- sapply(xreg2obj, function(x) sum(x$full$internal_count))
  oout$counts <- c(oout$counts, total = sum(unlist(oout$counts)))
  oout$valuevars <- lapply(xreg2obj, function(x) x$valuevar)
  oout$coef   <- coef

  oout$coef_wfixed <- fullcoef <- data.frame(Estimate = c(coef, oout$fixed))

  if (NROW(fullcoef)) {
    fullcoef[, "Std. Error"] <- NA
    fullcoef$type <- rep(c("Fitted", "Fixed"), c(ncoef, nfixed))
    if (hessian) fullcoef[seq_len(ncoef), "Std. Error"] <- oout$SE
  }

  oout$full_coef  <- fullcoef
  oout$method     <- method
  oout$optim_time <- Sys.time() - optim_time_start
  class(oout)     <- "xreg2"
  oout
}

#' Create an xreg2 control object
#'
#' @title Create xreg2 Control Object
#' @description Creates a control object specifying the model formula(s),
#'   likelihood function, gradient function, and starting values for one
#'   component of a joint \code{xreg2} model.
#'
#' @param formulas A formula or list of formulas. The left-hand side names the
#'   dependent (or intermediate) variable; the right-hand side is evaluated in
#'   the context of parameter values and data columns.
#' @param start_values Named numeric vector of starting values. An existing
#'   \code{xreg2} fitted object may be supplied; coefficients are extracted.
#' @param fixed_values Named numeric vector of parameter values held fixed
#'   during optimisation.
#' @param p_fun Log-likelihood function. Default \code{xreg2_cont_normal}.
#'   Must accept \code{(par, dtalist, ...)} and return a scalar.
#' @param grad_fun Gradient function. Default uses \code{numDeriv::grad} with
#'   method \code{"simple"}.
#' @param p_aggregation_fun Aggregation function applied to the likelihood
#'   data-frame. Rarely changed by users.
#' @param weights_var Character name of a column in the data used as
#'   observation weights. \code{NA} for unweighted.
#' @param name Optional character label for this component.
#' @param censor_bounds Numeric vector of length 2 giving lower and upper
#'   censoring bounds. Default \code{c(-Inf, Inf)} (no censoring).
#' @param lower Named numeric vector of lower parameter bounds (for
#'   bounded optimisers).
#' @param upper Named numeric vector of upper parameter bounds.
#'
#' @return An object of class \code{c("xreg2Control", "list")}.
#'
#' @seealso \code{\link{xreg2_fit}}, \code{\link{c.xreg2Control}}
#'
#' @examples
#' ctrl <- xreg2_control(
#'   formulas     = list(y ~ INTERCEPT + x * SLOPE),
#'   start_values = c(INTERCEPT = 0, SLOPE = 0),
#'   p_fun        = xreg2_cont_normal,
#'   name         = "continuous"
#' )
#'
#' @export
xreg2_control <- function(formulas,
                           start_values        = numeric(),
                           fixed_values        = numeric(),
                           p_fun               = xreg2_cont_normal,
                           grad_fun            = function(par, ...) numDeriv::grad(func = p_fun, x = par, method = "simple", ...),
                           p_aggregation_fun   = function(d_df) return(-d_df$p * d_df$internal_count),
                           weights_var         = NA,
                           name                = NA,
                           censor_bounds       = c(-Inf, Inf),
                           lower               = NA,
                           upper               = NA) {

  if (class(start_values)[1] %in% c("xreg", "xreg2")) start_values <- start_values$coef
  if (class(formulas)[1] %in% c("xreg2Control", "xreg2ControlList"))
    return(c(formulas))
  if (class(formulas)[1] == "formula")
    formulas <- list(formulas)

  required <- attr(x = p_fun, which = "required")

  defined_vars <- character()
  for (formula in formulas) {
    targetName   <- as.character(formula[[2]])
    defined_vars <- c(defined_vars, targetName)
  }

  if (length(required)) {
    req <- as.list(rep(0, length(required)))
    names(req) <- names(required)
    for (i in seq_along(required)) {
      if (!names(req)[i] %in% defined_vars) {
        defined_vars <- c(defined_vars, names(req)[i])
        req_return   <- required[[i]]
        formulas     <- c(formulas, list(req_return$formula))
        if (exists("start_values", req_return))
          start_values <- c(start_values,
                            req_return$start_values[!req_return$start_values %in% start_values])
        if (exists("fixed_values", req_return))
          fixed_values <- c(fixed_values,
                            req_return$fixed_values[!req_return$fixed_values %in% fixed_values])
      }
    }
  }

  start_values <- start_values[unique(names(start_values))]
  retObj <- mget(c(names(formals()), "defined_vars"))
  retObj$columnNames <- character()
  for (formula in formulas) {
    retObj$columnNames <- c(retObj$columnNames,
                            all.vars(formula)[!all.vars(formula) %in% retObj$columnNames])
  }
  retObj$columnNames <- retObj$columnNames[!retObj$columnNames %in% names(start_values)]
  class(retObj) <- c("xreg2Control", "list")
  return(retObj)
}

#' Combine xreg2Control objects into a list
#'
#' @title Combine xreg2 Control Objects
#' @description Combines one or more \code{xreg2Control} objects into an
#'   \code{xreg2ControlList} suitable for passing to \code{\link{xreg2_fit}}.
#'
#' @param ... One or more \code{xreg2Control} objects. Named arguments are
#'   used as component names; unnamed arguments use the object's \code{name}
#'   field or a sequential integer.
#'
#' @return An object of class \code{c("xreg2ControlList", "list")}.
#'
#' @seealso \code{\link{xreg2_control}}, \code{\link{xreg2_fit}}
#'
#' @examples
#' ctrl1 <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#'                         name = "component1")
#' cl    <- c(ctrl1)
#'
#' @export
c.xreg2Control <- function(...) {
  args <- list(...)

  if (is.null(names(args))) names(args) <- rep(NA_character_, length(args))
  for (i in seq_along(args)) {
    if (is.na(names(args)[i])) {
      if (is.na(args[[i]]$name)) {
        names(args)[i] <- as.character(i)
      } else {
        names(args)[i] <- args[[i]]$name
      }
    }
    if (!"xreg2Control" %in% class(args[[i]]))
      stop("At least one non-xreg2Control object among arguments.")
  }
  class(args) <- c("xreg2ControlList", "list")

  columnNames <- character()
  startValues <- numeric()
  fixedValues <- numeric()
  lower       <- numeric()
  upper       <- numeric()

  for (xreg2Control in args) {
    columnNames <- c(columnNames,
                     xreg2Control$columnNames[!xreg2Control$columnNames %in% columnNames])
    startValues <- c(startValues,
                     xreg2Control$start_values[!names(xreg2Control$start_values) %in% names(startValues)])
    fixedValues <- c(fixedValues,
                     xreg2Control$fixed_values[!names(xreg2Control$fixed_values) %in% names(fixedValues)])
    lower <- c(lower, xreg2Control$lower[!xreg2Control$lower %in% lower])
    upper <- c(upper, xreg2Control$upper[!xreg2Control$upper %in% upper])
  }

  attr(args, which = "columnNames") <- columnNames
  attr(args, which = "startValues") <- startValues
  attr(args, which = "fixedValues") <- fixedValues
  attr(args, which = "lower")       <- lower
  attr(args, which = "upper")       <- upper

  return(args)
}

#' @rdname c.xreg2Control
#' @export
c.xreg2ControlList <- function(...) {
  args <- list(...)
  for (narg in seq_along(args)) {
    class(args[[narg]]) <- "list"
  }
  return(do.call(c.xreg2Control, do.call(c, args)))
}

# Internal utility functions for xreg2

# Safe log: replace non-positive values with machine epsilon before taking log
Log <- function(x) {
  x[x <= 0] <- .Machine$double.xmin
  log(x)
}

# Outer product wrapper
Outer <- function(y, x, FUN = "*") outer(y, x, FUN)

# Grouped sum: returns vector of length = number of unique groups
group.sum <- function(x, ina) as.numeric(tapply(x, ina, sum))

# Repeat a row vector n times to form an n x length(m) matrix
rep_row <- function(m, n) matrix(as.numeric(m), nrow = n, ncol = length(m), byrow = TRUE)

# Filter ... args to those accepted by fun
match_formals <- function(fun, ...) {
  passedargs <- list(...)
  passedargs[!names(passedargs) %in% names(formals(fun))] <- NULL
  passedargs
}

# Match rows across two data.frames with the same number of columns
multimatch <- function(df1, df2) {
  if (ncol(df1) != ncol(df2)) stop("Data.frames must have same number of columns.")
  a <- apply(df1, 1, function(x) paste(x, collapse = "_"))
  b <- apply(df2, 1, function(x) paste(x, collapse = "_"))
  return(match(a, b))
}

# Change left-hand side of a formula
change_target <- function(object, newtarget, oldtarget = NULL) UseMethod("change_target")

#' @exportS3Method
change_target.formula <- function(object, newtarget, oldtarget = NULL) {
  if (length(as.list(object)) == 2) {
    retval <- as.formula(paste0(newtarget, "~", as.character(object[2])))
  } else {
    target <- object[[2]]
    if (is.null(oldtarget)) oldtarget <- target
    if (oldtarget != target) return(object)
    retval <- as.formula(paste0(newtarget, "~", as.character(object[3])))
  }
  attr(retval, ".Environment") <- attr(object, ".Environment")
  return(retval)
}

# Prepend or append to the right-hand side of a formula
appendmodel <- function(object, prepend = "", append = "") UseMethod("appendmodel")

#' @exportS3Method
appendmodel.formula <- function(object, prepend = "", append = "") {
  LHS <- ""
  RHS <- as.character(object[length(object)])
  if (length(object) == 3) LHS <- as.character(object[2])
  RHS <- paste(prepend, RHS, append)
  return(formula(paste(LHS, " ~ ", RHS), env = environment(object)))
}

# Safe try: return NULL on error
try0 <- function(expr) tryCatch(expr, error = function(e) NULL)

test_that("internal helpers work correctly", {
  # Log: safe log
  expect_equal(Log(0), log(.Machine$double.xmin))
  expect_equal(Log(1), 0)
  expect_true(is.finite(Log(-1)))

  # Outer: wrapper around outer()
  result <- Outer(1:3, 1:3, "+")
  expect_equal(dim(result), c(3L, 3L))
  expect_equal(result[1, 1], 2)
  expect_equal(result[2, 3], 5)

  # group.sum: grouped sum
  x   <- c(1, 2, 3, 4, 5)
  ina <- c(1, 1, 2, 2, 2)
  gs  <- group.sum(x, ina)
  expect_equal(gs, c(3, 12))

  # rep_row: repeat row vector
  m <- c(1, 2, 3)
  r <- rep_row(m, 4)
  expect_equal(dim(r), c(4L, 3L))
  expect_true(all(r[2, ] == m))

  # match_formals: filter to accepted args
  f   <- function(a, b, c = 1) a + b + c
  mf  <- match_formals(f, a = 10, b = 20, d = 99)
  expect_true("a" %in% names(mf))
  expect_true("b" %in% names(mf))
  expect_false("d" %in% names(mf))
})

test_that("xreg2_control creates correct class", {
  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 0, SLOPE = 0),
    name         = "test"
  )
  expect_s3_class(ctrl, "xreg2Control")
  expect_true("xreg2Control" %in% class(ctrl))
  # After c(), we get an xreg2ControlList
  cl <- c(ctrl)
  expect_s3_class(cl, "xreg2ControlList")
})

test_that("xreg2_control adds required sigma formula for xreg2_cont_normal", {
  ctrl <- xreg2_control(
    formulas = list(y ~ INTERCEPT + x * SLOPE),
    p_fun    = xreg2_cont_normal,
    name     = "test"
  )
  # sigma_est formula should be added automatically from required attr
  formula_targets <- sapply(ctrl$formulas, function(f) as.character(f[[2]]))
  expect_true("sigma_est" %in% formula_targets)
  # LN_SIGMA should be in start_values
  expect_true("LN_SIGMA" %in% names(ctrl$start_values))
})

test_that("xreg2_fit produces finite coefficients on synthetic data", {
  skip_if_not_installed("ucminf")
  set.seed(42)
  n   <- 150
  df  <- data.frame(y = rnorm(n, 2, 0.5), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 2, SLOPE = 0),
    name         = "main"
  )
  cl  <- c(ctrl)
  fit <- xreg2_fit(controlList = cl, dataList = list(main = df))

  expect_s3_class(fit, "xreg2")
  expect_true(all(is.finite(fit$coef)))
  expect_true(is.finite(fit$minima[["total"]]))
  # Intercept should be close to 2
  expect_equal(fit$coef[["INTERCEPT"]], 2, tolerance = 0.2)
})

test_that("predict.xreg2 returns augmented data frame", {
  skip_if_not_installed("ucminf")
  set.seed(1)
  n   <- 100
  df  <- data.frame(y = rnorm(n, 1, 0.3), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 1, SLOPE = 0),
    name         = "main"
  )
  fit  <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))

  newdf <- data.frame(x = c(-1, 0, 1))
  preds <- predict(fit, newdata = list(main = newdf))

  expect_type(preds, "list")
  expect_true("main" %in% names(preds))
  expect_true("Xb" %in% colnames(preds[["main"]]))
  expect_equal(nrow(preds[["main"]]), 3)
})

test_that("AIC and BIC return finite numeric values", {
  skip_if_not_installed("ucminf")
  set.seed(7)
  n   <- 100
  df  <- data.frame(y = rnorm(n, 0, 1), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 0, SLOPE = 0),
    name         = "main"
  )
  fit <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))

  aic_val <- AIC(fit)
  bic_val <- BIC(fit)

  expect_true(is.numeric(aic_val))
  expect_true(is.finite(aic_val))
  expect_true(is.numeric(bic_val))
  expect_true(is.finite(bic_val))
  # BIC >= AIC for n > e^2 ~ 7.4
  expect_true(bic_val >= aic_val)
})

test_that("xreg2_cont_normal attr is set correctly", {
  req <- attr(xreg2_cont_normal, "required")
  expect_type(req, "list")
  expect_true("sigma_est" %in% names(req))
  expect_true("LN_SIGMA" %in% names(req$sigma_est$start_values))
})

test_that("xreg2_cont_r_normal attr is set correctly", {
  req <- attr(xreg2_cont_r_normal, "required")
  expect_type(req, "list")
  expect_true("sigma_est" %in% names(req))
  expect_true("omega_est" %in% names(req))
})

test_that("L-BFGS-B method produces finite coefficients", {
  set.seed(42)
  n  <- 150
  df <- data.frame(y = rnorm(n, 2, 0.5), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 2, SLOPE = 0),
    name         = "main"
  )
  fit <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
                   method = "L-BFGS-B")

  expect_s3_class(fit, "xreg2")
  expect_true(all(is.finite(fit$coef)))
  expect_equal(fit$method, "L-BFGS-B")
  # Intercept should be close to the data mean (~2)
  expect_equal(fit$coef[["INTERCEPT"]], 2, tolerance = 0.2)
})

test_that("L-BFGS-B respects lower bound on INTERCEPT", {
  set.seed(10)
  n  <- 200
  df <- data.frame(y = rnorm(n, 0.5, 0.3), x = rnorm(n))

  # Constrain INTERCEPT >= 0.8 (above the true mean of 0.5),
  # so the optimizer should be pushed to the bound.
  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 1, SLOPE = 0),
    lower        = c(INTERCEPT = 0.8),
    name         = "main"
  )
  fit <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
                   method = "L-BFGS-B")

  expect_s3_class(fit, "xreg2")
  expect_true(all(is.finite(fit$coef)))
  # Bound must be respected
  expect_gte(fit$coef[["INTERCEPT"]], 0.8 - 1e-6)
})

test_that("L-BFGS-B respects upper bound on INTERCEPT", {
  set.seed(20)
  n  <- 200
  df <- data.frame(y = rnorm(n, 3, 0.3), x = rnorm(n))

  # Constrain INTERCEPT <= 2 (below the true mean of 3).
  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 2, SLOPE = 0),
    upper        = c(INTERCEPT = 2),
    name         = "main"
  )
  fit <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
                   method = "L-BFGS-B")

  expect_s3_class(fit, "xreg2")
  expect_true(all(is.finite(fit$coef)))
  expect_lte(fit$coef[["INTERCEPT"]], 2 + 1e-6)
})

test_that("optim_control overrides default L-BFGS-B settings", {
  set.seed(5)
  n  <- 100
  df <- data.frame(y = rnorm(n, 1, 0.5), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 1, SLOPE = 0),
    name         = "main"
  )
  # maxit = 1 forces early termination; fit should still complete without error
  fit <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
                   method = "L-BFGS-B",
                   optim_control = list(maxit = 1))

  expect_s3_class(fit, "xreg2")
})

test_that("unsupported method without optimx gives informative error", {
  skip_if(requireNamespace("optimx", quietly = TRUE),
          "optimx is installed; skipping fallback-error test")
  set.seed(1)
  n  <- 50
  df <- data.frame(y = rnorm(n), x = rnorm(n))

  ctrl <- xreg2_control(
    formulas     = list(y ~ INTERCEPT + x * SLOPE),
    start_values = c(INTERCEPT = 0, SLOPE = 0),
    name         = "main"
  )
  expect_error(
    xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
              method = "nlminb"),
    regexp = "optimx"
  )
})

#' Predict method for xreg2 objects
#'
#' @title Predict from an xreg2 Model
#' @description Generates predictions (linear predictors and intermediate
#'   computed variables) from a fitted \code{xreg2} model applied to new data.
#'
#' @param object A fitted \code{xreg2} object.
#' @param newdata A named list of data frames (one per model component), or a
#'   single data frame (automatically wrapped and named after the first
#'   component).
#' @param return_vector Logical; if \code{TRUE} and there is a single
#'   component, return a numeric vector of \code{Xb} values instead of a
#'   list. Default \code{FALSE}.
#' @param ... Currently unused.
#'
#' @return A named list of data frames (one per component), each augmented
#'   with computed variables including \code{Xb}.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
#' ctrl <- xreg2_control(
#'   formulas     = list(y ~ INTERCEPT + x * SLOPE),
#'   start_values = c(INTERCEPT = 0, SLOPE = 0),
#'   name         = "main"
#' )
#' fit  <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))
#' newdf <- data.frame(x = 1:5)
#' preds <- predict(fit, newdata = list(main = newdf))
#' }
#'
#' @export
predict.xreg2 <- function(object, newdata = NULL, return_vector = FALSE, ...) {
  if (is.data.frame(newdata)) {
    newdata <- list(newdata)
    names(newdata) <- names(object$formulas)[1]
  }
  if (!all(names(newdata) %in% names(object$formulas)))
    stop("Data type listed in newdata not in xreg2 object.")

  args <- as.list(object$parf)
  dtns <- names(newdata)
  names(dtns) <- dtns

  preds <- lapply(dtns, function(dtn) {
    formulas <- object$formulas[[dtn]]
    d_df     <- newdata[[dtn]]
    for (formulanum in seq_along(formulas)) {
      formula <- formulas[[formulanum]]
      new_var <- as.character(formula[[2]])
      if (new_var %in% c(colnames(d_df), object$valuevars[[dtn]])) new_var <- "Xb"
      d_df[, new_var] <- with(args, with(d_df, eval(parse(text = formula))))
    }
    d_df
  })

  if (length(newdata) == 1 && return_vector) return(preds[[1]]$Xb)
  return(preds)
}

#' Print method for xreg2 objects
#'
#' @title Print an xreg2 Model
#' @description Prints a concise summary of a fitted \code{xreg2} model,
#'   including coefficients, standard errors, and minimised negative
#'   log-likelihoods.
#'
#' @param x A fitted \code{xreg2} object.
#' @param ... Currently unused.
#'
#' @return \code{x} invisibly.
#'
#' @export
print.xreg2 <- function(x, ...) {
  cat("\nxreg2 fitted model\n")
  cat("Method:", x$method, "\n")
  cat("Optimisation time:", format(x$optim_time), "\n\n")

  if (NROW(x$full_coef)) {
    cat("Coefficients:\n")
    print(x$full_coef)
  }

  cat("\nNegative log-likelihoods:\n")
  print(x$minima)

  cat("\nObservation counts:\n")
  print(x$counts)

  invisible(x)
}

#' Summary method for xreg2 objects
#'
#' @title Summarise an xreg2 Model
#' @description Prints a detailed summary of a fitted \code{xreg2} model.
#'   Currently equivalent to \code{\link{print.xreg2}}.
#'
#' @param object A fitted \code{xreg2} object.
#' @param ... Currently unused.
#'
#' @return \code{object} invisibly.
#'
#' @export
summary.xreg2 <- function(object, ...) {
  print(object, ...)
}

#' AIC method for xreg2 objects
#'
#' @title AIC for an xreg2 Model
#' @importFrom stats AIC BIC
#' @description Computes the Akaike Information Criterion for a fitted
#'   \code{xreg2} model. Scale parameters (SIGMA, LN_SIGMA, OMEGA, LN_OMEGA)
#'   are excluded from the parameter count.
#'
#' @param object A fitted \code{xreg2} object.
#' @param ... Currently unused.
#' @param k Numeric; penalty per parameter. Default \code{2}.
#' @param just_total Logical; return only the total AIC across all components.
#'   Default \code{TRUE}.
#'
#' @return A numeric scalar (AIC value).
#'
#' @export
AIC.xreg2 <- function(object, ..., k = 2, just_total = TRUE) {
  coefs <- object$coef[!names(object$coef) %in%
                         c("SIGMA", "SIGMA_INTERCEPT", "LN_SIGMA", "OMEGA", "LN_OMEGA")]
  npar  <- length(coefs)
  mini  <- object$minima
  if (just_total) mini <- mini[["total"]]
  return(2 * mini + k * npar)
}

#' BIC method for xreg2 objects
#'
#' @title BIC for an xreg2 Model
#' @description Computes the Bayesian Information Criterion for a fitted
#'   \code{xreg2} model.
#'
#' @param object A fitted \code{xreg2} object.
#' @param ... Currently unused.
#' @param just_total Logical; return only the total BIC. Default \code{TRUE}.
#'
#' @return A numeric scalar (BIC value).
#'
#' @export
BIC.xreg2 <- function(object, ..., just_total = TRUE) {
  counts <- object$counts
  names(counts)[length(counts)] <- "total"
  if (just_total) counts <- counts[["total"]]
  AIC(object, ..., k = log(counts), just_total = just_total)
}

#' Hybrid regression wrapper for xreg2
#'
#' @title Hybrid Regression (xreg2 version)
#' @description High-level wrapper around \code{\link{xreg2_fit}} replicating
#'   the interface of the STATA \code{hyreg} command and
#'   \code{xreg::hyreg()}. Simultaneously fits a censored continuous model
#'   (normal or logistic) and a dichotomous model (logistic or normal) to a
#'   single data frame that contains both observation types.
#'
#' @param formula A formula or list of formulas for the continuous component.
#'   The dependent variable (left-hand side) must match a column in \code{df},
#'   or use interval columns named \code{varname.lb} and \code{varname.ub}.
#' @param df A \code{data.frame} containing both continuous and dichotomous
#'   observations. Must include the column specified by \code{datatype}.
#' @param datatype Character; name of the column that distinguishes continuous
#'   (\code{TRUE}) from dichotomous (\code{FALSE}) observations. Default
#'   \code{"method"}.
#' @param init Optional named numeric vector (or \code{xreg2} object) of
#'   starting values.
#' @param contdist Character; distribution for the continuous component.
#'   \code{"normal"} (default) or \code{"logistic"}.
#' @param dichdist Character; distribution for the dichotomous component.
#'   \code{"logistic"} (default) or \code{"normal"}.
#' @param hetcont Optional formula for heteroscedastic standard deviations in
#'   the continuous component. If \code{NULL} (default), uses
#'   \code{sigma_est ~ SIGMA} (or \code{sigma_est ~ exp(SIGMA)} when
#'   \code{lnsigma = TRUE}).
#' @param hetdich Optional formula for heteroscedastic standard deviations in
#'   the dichotomous component. If \code{NULL} (default), uses
#'   \code{theta_est ~ THETA} (or \code{theta_est ~ exp(THETA)} when
#'   \code{lntheta = TRUE}).
#' @param ul Numeric; upper censoring bound. Default \code{Inf}.
#' @param ll Numeric; lower censoring bound. Default \code{-Inf}.
#' @param lntheta Logical; exponentiate theta prior to fitting (matches STATA
#'   default). Default \code{TRUE}.
#' @param lnsigma Logical; exponentiate sigma prior to fitting (matches STATA
#'   default). Default \code{TRUE}.
#' @param dichformula Optional formula for the dichotomous component. Defaults
#'   to \code{formula}.
#' @param ... Additional arguments forwarded to \code{\link{xreg2_fit}}.
#'
#' @return An object of class \code{"xreg2"}.
#'
#' @seealso \code{\link{xreg2_fit}}, \code{\link{xreg2_control}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' n   <- 300
#' df  <- data.frame(
#'   value  = c(rnorm(200, 0.6, 0.3), rep(NA, 100)),
#'   age    = rnorm(n),
#'   method = c(rep(TRUE, 200), rep(FALSE, 100))
#' )
#' df$value[df$method == FALSE] <- as.integer(df$age[df$method == FALSE] > 0)
#'
#' fit <- xreg2_hyreg(
#'   formula  = value ~ INTERCEPT + age * AGE,
#'   df       = df,
#'   datatype = "method"
#' )
#' print(fit)
#' }
#'
#' @export
xreg2_hyreg <- function(formula,
                          df,
                          datatype     = "method",
                          init         = numeric(),
                          contdist     = "normal",
                          dichdist     = "logistic",
                          hetcont      = NULL,
                          hetdich      = NULL,
                          ul           = Inf,
                          ll           = -Inf,
                          lntheta      = TRUE,
                          lnsigma      = TRUE,
                          dichformula  = NULL,
                          ...) {

  if (!"data.frame" %in% class(df))
    stop(paste0(class(df)[1], " provided to the df argument. Should be data.frame."))
  if (!datatype %in% colnames(df))
    stop(paste0("No column named \"", datatype, "\" in provided dataframe."))
  if (!all(as.logical(unique(df[, datatype])) %in% c(TRUE, FALSE)))
    stop("Datatype column in provided data.frame must be interpretable as boolean.")
  if (!contdist %in% c("normal", "logistic"))
    stop('contdist-argument has to be either "normal" or "logistic".')
  if (!dichdist %in% c("normal", "logistic"))
    stop('dichdist-argument has to be either "normal" or "logistic".')

  if (is.null(dichformula)) dichformula <- formula
  if (inherits(formula,     "formula")) formula     <- list(formula)
  if (inherits(dichformula, "formula")) dichformula <- list(dichformula)

  dataList <- list(
    Continuous  = df[as.logical(df[, datatype] == TRUE),  ],
    Dichotomous = df[as.logical(df[, datatype] == FALSE), ]
  )

  dot_args <- list(...)
  dot_args[c("p_fun", "p_aggrgation_fun")] <- NULL

  if (!is.null(hetcont)) {
    if (is.character(hetcont)) {
      tmp     <- try0(as.formula(hetcont))
      if (is.null(tmp)) tmp <- try0(as.formula(paste0("~ ", hetcont)))
      if (is.null(tmp)) stop("Could not generate useful formula from provided hetcont-argument.")
      hetcont <- tmp
    }
    hetcont <- change_target(hetcont, "sigma_est")
  } else {
    hetcont <- as.formula("sigma_est ~ SIGMA")
  }

  if (!is.null(hetdich)) {
    if (is.character(hetdich)) {
      tmp     <- try0(as.formula(hetdich))
      if (is.null(tmp)) tmp <- try0(as.formula(paste0("~ ", hetdich)))
      if (is.null(tmp)) stop("Could not generate useful formula from provided hetdich-argument.")
      hetdich <- tmp
    }
    hetdich <- change_target(hetdich, "theta_est")
  } else {
    hetdich <- as.formula("theta_est ~ THETA")
  }

  if (lnsigma)  hetcont <- appendmodel(hetcont, " exp(", ")")
  if (lntheta)  hetdich <- appendmodel(hetdich, " exp(", ")")

  cont_formula <- c(formula, list(hetcont))
  dich_formula <- c(dichformula, list(hetdich))

  cont_p_fun <- xreg2_cont_normal
  dich_p_fun <- xreg2_dich_logistic
  if (contdist == "logistic") cont_p_fun <- xreg2_dich_logistic
  if (dichdist == "normal")   dich_p_fun <- xreg2_cont_normal

  cont_control <- xreg2_control(formulas = cont_formula, p_fun = cont_p_fun,
                                 name = "Continuous",   censor_bounds = c(ll, ul))
  dich_control <- xreg2_control(formulas = dich_formula, p_fun = dich_p_fun,
                                 name = "Dichotomous")
  control <- c(cont_control, dich_control)

  ret <- do.call("xreg2_fit", c(
    list(controlList  = control,
         dataList     = dataList,
         start_values = init,
         run_from     = "hyreg"),
    dot_args
  ))

  return(ret)
}

# Likelihood functions and gradient functions for xreg2

# ---------------------------------------------------------------------------
# Internal: negative log-likelihood dispatcher
# ---------------------------------------------------------------------------

negll_list2 <- function(par, dtalist, ...) {
  sum(sapply(dtalist, function(dtl) {
    do.call(dtl$ll_fun, args = c(list(dtalist = dtl, par = par), list(...)))
  }))
}

# Internal: gradient dispatcher
gradll_list2 <- function(par, dtalist, ...) {
  rowSums(as.matrix(sapply(dtalist, function(dtl) {
    do.call(dtl$grad_fun, args = c(list(dtalist = dtl, par = par), list(...)))
  })))
}

# ---------------------------------------------------------------------------
# Internal: forward gradient for continuous normal
# ---------------------------------------------------------------------------

grad_cont_normal2 <- function(par, dtalist, ret_mat = FALSE, eps = 1e-7, centered = TRUE) {

  if (centered) par <- par - eps / 2

  d_df <- dtalist$unique
  f_df <- dtalist$full
  nobs <- NROW(d_df)

  names(par) <- dtalist$coeff_names
  plen  <- length(par)
  plen1 <- plen + 1

  parm <- as.data.frame(as.list(par))
  parm <- parm[rep(1, plen + 1), ]

  for (pn in 1:plen) {
    parm[pn + 1, pn] <- parm[pn + 1, pn] + eps
  }
  par <- parm

  formulas <- dtalist$formulas

  for (formulanum in seq_along(formulas)) {
    formula  <- formulas[[formulanum]]
    new_var  <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    tmp <- sapply(1:plen1, function(x) {
      theseargs <- c(
        as.list(par),
        as.list(dtalist$fixed),
        as.list(d_df[, colnames(d_df)[colnames(d_df) %in% all.vars(formula)], drop = FALSE])
      )
      theseargs <- lapply(theseargs, function(thisarg) {
        if (!is.null(dim(thisarg)) && dim(thisarg)[2] == plen1) return(thisarg[, x])
        if (length(thisarg) == plen1) return(thisarg[x])
        thisarg
      })
      with(theseargs, eval(parse(text = formula)))
    })
    if (is.null(dim(tmp)) && length(tmp) == plen1)
      tmp <- matrix(tmp, nrow = nobs, ncol = plen1, byrow = TRUE)
    d_df[, new_var] <- tmp
  }

  pvals <- matrix(NA, nrow = nobs, ncol = plen1)

  sel  <- d_df$internal_type == 1
  pvals[sel, ]  <- dnorm(x    = d_df$Xb[sel, ],
                         mean = d_df$internal_lb[sel],
                         sd   = d_df$sigma_est[sel, ],
                         log  = TRUE)
  sel <- !sel
  pvals[sel, ] <- log(
    pnorm(q = d_df$Xb[sel, ], mean = d_df$internal_lb[sel],
          sd = d_df$sigma_est[sel, ], lower.tail = TRUE,  log.p = FALSE) -
    pnorm(q = d_df$Xb[sel, ], mean = d_df$internal_ub[sel],
          sd = d_df$sigma_est[sel, ], lower.tail = TRUE,  log.p = FALSE)
  )

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  pvals <- pvals[f_df$umatch, ] * f_df$internal_count
  if (ret_mat) return(pvals)
  pvals <- colSums(pvals)
  pvals <- (pvals[1] - pvals[2:plen1]) / eps
  return(pvals)
}

# ---------------------------------------------------------------------------
# Internal: forward gradient for dichotomous logistic
# ---------------------------------------------------------------------------

grad_dich_logistic2 <- function(par, dtalist, ret_mat = FALSE, eps = 1e-7, centered = TRUE) {

  if (centered) par <- par - eps / 2

  d_df <- dtalist$unique
  f_df <- dtalist$full
  nobs <- NROW(d_df)

  names(par) <- dtalist$coeff_names
  plen  <- length(par)
  plen1 <- plen + 1

  parm <- as.data.frame(as.list(par))
  parm <- parm[rep(1, plen + 1), ]

  for (pn in 1:plen) {
    parm[pn + 1, pn] <- parm[pn + 1, pn] + eps
  }
  par <- parm

  formulas <- dtalist$formulas

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    tmp <- sapply(1:plen1, function(x) {
      theseargs <- c(as.list(par[x, ]), as.list(dtalist$fixed))
      with(theseargs, with(d_df, eval(parse(text = formula))))
    })
    if (is.null(dim(tmp)) && length(tmp) == plen1)
      tmp <- matrix(tmp, nrow = nobs, ncol = plen1, byrow = TRUE)
    d_df[, new_var] <- tmp
  }

  Xb <- d_df$Xb
  if ("theta_est" %in% colnames(d_df)) Xb <- Xb * d_df$theta_est

  logistic_tmp <- .5 + .5 * tanh(Xb / 2)
  pvals <- log(d_df$internal_lb *      logistic_tmp +
               (1 - d_df$internal_lb) * (1 - logistic_tmp))
  pvals <- matrix(pvals, nrow = nobs, ncol = plen1)

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  pvals <- pvals[f_df$umatch, ] * f_df$internal_count
  if (ret_mat) return(pvals)
  pvals <- colSums(pvals)
  pvals <- (pvals[1] - pvals[2:plen1]) / eps
  return(pvals)
}

# ---------------------------------------------------------------------------
# Exported: continuous normal log-likelihood (fast list-based implementation)
# ---------------------------------------------------------------------------

#' Continuous normal log-likelihood for xreg2
#'
#' @title Continuous Normal Negative Log-Likelihood
#' @description Computes the negative log-likelihood for a censored/interval
#'   normal regression model. Designed as a \code{p_fun} argument for
#'   \code{\link{xreg2_control}}.
#'
#' @param par Named numeric vector of current parameter values.
#' @param dtalist Internal data list produced by \code{xreg2_fit}.
#' @param log.p Logical; use log-scale probabilities. Default \code{TRUE}.
#' @param aggregate.p Logical; return scalar negloglik. Default \code{TRUE}.
#' @param print_pars Logical; print parameters and value. Default \code{FALSE}.
#'
#' @return Scalar negative log-likelihood (when \code{aggregate.p = TRUE}).
#'
#' @examples
#' # Typically used as p_fun in xreg2_control():
#' # ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#' #                        p_fun = xreg2_cont_normal)
#'
#' @export
xreg2_cont_normal <- function(par, dtalist, log.p = TRUE, aggregate.p = TRUE,
                               print_pars = FALSE) {
  names(par) <- dtalist$coeff_names
  if (print_pars)
    message(do.call(paste, lapply(names(par),
                                  function(x) paste0(x, ": ", round(par[[x]], 3)))))

  d_l  <- as.list(dtalist$unique)
  nobs <- length(d_l[[1]])

  lapply(dtalist$formulas, function(formula) {
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(names(d_l), dtalist$valuevar)) new_var <- "Xb"
    d_l[[new_var]] <<- rep_len(
      eval(formula[[3]], envir = c(as.list(c(par, dtalist$fixed)), d_l)),
      nobs)
    NULL
  })

  logf  <- if (log.p) function(x) log(x) else function(x) x
  pvals <- matrix(NA, nrow = nobs, ncol = 1)

  sel <- which(d_l$internal_type == 1)
  pvals[sel, ] <- dnorm(x    = d_l$internal_lb[sel],
                        mean = d_l$Xb[sel],
                        sd   = d_l$sigma_est[sel],
                        log  = log.p)

  sel <- which(d_l$internal_type != 1)
  pvals[sel, ] <- logf(
    pnorm(q = d_l$internal_lb[sel], mean = d_l$Xb[sel],
          sd = d_l$sigma_est[sel], lower.tail = FALSE, log.p = FALSE) -
    pnorm(q = d_l$internal_ub[sel], mean = d_l$Xb[sel],
          sd = d_l$sigma_est[sel], lower.tail = FALSE, log.p = FALSE)
  )

  if (!aggregate.p) return(pvals)

  if (log.p) {
    if ("boundNA" %in% names(dtalist))
      if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
    pvals[pvals == -Inf] <- log(.Machine$double.xmin)
    pvals <- -sum(pvals[dtalist$full$umatch, ] * dtalist$full$internal_count)
  } else {
    if ("boundNA" %in% names(dtalist))
      if (dtalist$boundNA) pvals[is.na(pvals)] <- .Machine$double.xmin
    pvals[pvals == 0] <- .Machine$double.xmin
    pvals <- -prod(pvals[dtalist$full$umatch, ] * dtalist$full$internal_count)
  }

  if (is.na(pvals)) {
    if (print_pars) message("Encountered NA")
    pvals <- 10e10
  }
  if (print_pars) message(paste0("Value: ", pvals))
  return(pvals)
}

attr(x = xreg2_cont_normal, which = "required") <- list(
  sigma_est = list(
    formula      = sigma_est ~ exp(LN_SIGMA),
    start_values = c(LN_SIGMA = 0)
  )
)

# ---------------------------------------------------------------------------
# Exported: general continuous log-likelihood (multi-distribution)
# ---------------------------------------------------------------------------

#' General continuous log-likelihood for xreg2
#'
#' @title General Continuous Negative Log-Likelihood
#' @description Computes the negative log-likelihood for a censored/interval
#'   regression model supporting multiple distributions (normal, lognormal,
#'   exponential, beta, gamma, weibull). Designed as a \code{p_fun} argument
#'   for \code{\link{xreg2_control}}.
#'
#' @param par Named numeric vector of current parameter values.
#' @param dtalist Internal data list produced by \code{xreg2_fit}.
#' @param log.p Logical; use log-scale probabilities. Default \code{TRUE}.
#' @param aggregate.p Logical; return scalar negloglik. Default \code{TRUE}.
#' @param print_pars Logical; print parameters and value. Default \code{FALSE}.
#' @param dist Character; distribution family. One of \code{"normal"},
#'   \code{"lognormal"}, \code{"exp"}/\code{"exponential"}, \code{"beta"},
#'   \code{"gamma"}, \code{"weibull"}. Default \code{"normal"}.
#'
#' @return Scalar negative log-likelihood (when \code{aggregate.p = TRUE}).
#'
#' @examples
#' # ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#' #                        p_fun = xreg2_cont,
#' #                        start_values = c(INTERCEPT = 0, SLOPE = 0, LN_SIGMA = 0))
#'
#' @export
xreg2_cont <- function(par, dtalist, log.p = TRUE, aggregate.p = TRUE,
                        print_pars = FALSE, dist = "normal") {
  p_type <- dtalist$type
  names(par) <- dtalist$coeff_names

  formulas <- dtalist$formulas
  d_df <- dtalist$unique
  f_df <- dtalist$full

  nobs <- NROW(d_df)

  args <- c(as.list(par), as.list(dtalist$fixed))
  d_df <- cbind(d_df, args)

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    d_df[, new_var] <- with(d_df, eval(parse(text = formula)))
  }

  pvals <- matrix(NA, nrow = nobs, ncol = 1)
  sel   <- d_df$internal_type == 1

  if (dist == "normal") {
    pvals[sel, ] <- dnorm(x    = d_df$internal_lb[sel],
                          mean = d_df$Xb[sel],
                          sd   = d_df$sigma_est[sel],
                          log  = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        pnorm(q = d_df$internal_lb[sel], mean = d_df$Xb[sel],
              sd = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE) -
        pnorm(q = d_df$internal_ub[sel], mean = d_df$Xb[sel],
              sd = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        pnorm(q = d_df$internal_lb[sel], mean = d_df$Xb[sel],
              sd = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE) -
        pnorm(q = d_df$internal_ub[sel], mean = d_df$Xb[sel],
              sd = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE)
    }

  } else if (dist == "lognormal") {
    pvals[sel, ] <- dlnorm(x       = d_df$internal_lb[sel],
                           meanlog = d_df$Xb[sel],
                           sdlog   = d_df$sigma_est[sel],
                           log     = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        plnorm(q = d_df$internal_lb[sel], meanlog = d_df$Xb[sel],
               sdlog = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE) -
        plnorm(q = d_df$internal_ub[sel], meanlog = d_df$Xb[sel],
               sdlog = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        plnorm(q = d_df$internal_lb[sel], meanlog = d_df$Xb[sel],
               sdlog = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE) -
        plnorm(q = d_df$internal_ub[sel], meanlog = d_df$Xb[sel],
               sdlog = d_df$sigma_est[sel], lower.tail = FALSE, log.p = FALSE)
    }

  } else if (dist %in% c("exp", "exponential")) {
    d_df$Xb[sel & d_df$Xb < 0] <- 0
    pvals[sel, ] <- dexp(x    = d_df$internal_lb[sel],
                         rate = d_df$Xb[sel],
                         log  = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        pexp(q = d_df$internal_lb[sel], rate = d_df$Xb[sel],
             lower.tail = FALSE, log.p = FALSE) -
        pexp(q = d_df$internal_ub[sel], rate = d_df$Xb[sel],
             lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        pexp(q = d_df$internal_lb[sel], rate = d_df$Xb[sel],
             lower.tail = FALSE, log.p = FALSE) -
        pexp(q = d_df$internal_ub[sel], rate = d_df$Xb[sel],
             lower.tail = FALSE, log.p = FALSE)
    }

  } else if (dist == "beta") {
    d_df$Xb[sel & d_df$Xb < 0] <- 0
    pvals[sel, ] <- dbeta(x      = d_df$internal_lb[sel] * d_df$SCALE[sel],
                          shape1 = d_df$shape1[sel],
                          shape2 = d_df$shape2[sel],
                          log    = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        pbeta(q = d_df$internal_lb[sel] * d_df$SCALE[sel],
              shape1 = d_df$shape1[sel], shape2 = d_df$shape2[sel],
              lower.tail = FALSE, log.p = FALSE) -
        pbeta(q = d_df$internal_ub[sel] * d_df$SCALE[sel],
              shape1 = d_df$shape1[sel], shape2 = d_df$shape2[sel],
              lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        pbeta(q = d_df$internal_lb[sel] * d_df$SCALE[sel],
              shape1 = d_df$shape1[sel], shape2 = d_df$shape2[sel],
              lower.tail = FALSE, log.p = FALSE) -
        pbeta(q = d_df$internal_ub[sel] * d_df$SCALE[sel],
              shape1 = d_df$shape1[sel], shape2 = d_df$shape2[sel],
              lower.tail = FALSE, log.p = FALSE)
    }

  } else if (dist == "gamma") {
    d_df$Xb[sel & d_df$Xb < 0] <- 0
    pvals[sel, ] <- dgamma(x     = d_df$internal_lb[sel],
                           shape = d_df$shape[sel],
                           rate  = d_df$rate[sel],
                           log   = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        pgamma(q = d_df$internal_lb[sel], shape = d_df$shape[sel],
               rate = d_df$rate[sel], lower.tail = FALSE, log.p = FALSE) -
        pgamma(q = d_df$internal_ub[sel], shape = d_df$shape[sel],
               rate = d_df$rate[sel], lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        pgamma(q = d_df$internal_lb[sel], shape = d_df$shape[sel],
               rate = d_df$rate[sel], lower.tail = FALSE, log.p = FALSE) -
        pgamma(q = d_df$internal_ub[sel], shape = d_df$shape[sel],
               rate = d_df$rate[sel], lower.tail = FALSE, log.p = FALSE)
    }

  } else if (dist == "weibull") {
    d_df$Xb[sel & d_df$Xb < 0] <- 0
    pvals[sel, ] <- dweibull(x     = d_df$internal_lb[sel],
                             shape = d_df$shape[sel],
                             scale = d_df$scale[sel],
                             log   = log.p)
    sel <- !sel
    if (log.p) {
      pvals[sel, ] <- log(
        pweibull(q = d_df$internal_lb[sel], shape = d_df$shape[sel],
                 scale = d_df$scale[sel], lower.tail = FALSE, log.p = FALSE) -
        pweibull(q = d_df$internal_ub[sel], shape = d_df$shape[sel],
                 scale = d_df$scale[sel], lower.tail = FALSE, log.p = FALSE))
    } else {
      pvals[sel, ] <-
        pweibull(q = d_df$internal_lb[sel], shape = d_df$shape[sel],
                 scale = d_df$scale[sel], lower.tail = FALSE, log.p = FALSE) -
        pweibull(q = d_df$internal_ub[sel], shape = d_df$shape[sel],
                 scale = d_df$scale[sel], lower.tail = FALSE, log.p = FALSE)
    }
  }

  if (!aggregate.p) return(pvals)

  if (log.p) {
    if ("boundNA" %in% names(dtalist))
      if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
    pvals[pvals == -Inf] <- log(.Machine$double.xmin)
    pvals <- pvals[f_df$umatch, ]
    pvalso <- -sum(pvals * f_df$internal_count)
  } else {
    if ("boundNA" %in% names(dtalist))
      if (dtalist$boundNA) pvals[is.na(pvals)] <- .Machine$double.xmin
    pvals[pvals == 0] <- .Machine$double.xmin
    pvals <- pvals[f_df$umatch, ]
    pvalso <- -prod(pvals * f_df$internal_count)
  }
  return(pvalso)
}

# ---------------------------------------------------------------------------
# Exported: dichotomous logistic log-likelihood
# ---------------------------------------------------------------------------

#' Dichotomous logistic log-likelihood for xreg2
#'
#' @title Dichotomous Logistic Negative Log-Likelihood
#' @description Computes the negative log-likelihood for a binary logistic
#'   regression model. Designed as a \code{p_fun} argument for
#'   \code{\link{xreg2_control}}.
#'
#' @param par Named numeric vector of current parameter values.
#' @param dtalist Internal data list produced by \code{xreg2_fit}.
#'
#' @return Scalar negative log-likelihood.
#'
#' @examples
#' # ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#' #                        p_fun = xreg2_dich_logistic,
#' #                        start_values = c(INTERCEPT = 0, SLOPE = 0))
#'
#' @export
xreg2_dich_logistic <- function(par, dtalist) {
  p_type <- dtalist$type
  names(par) <- dtalist$coeff_names

  formulas <- dtalist$formulas
  d_df <- dtalist$unique
  f_df <- dtalist$full

  nobs <- NROW(d_df)

  args <- c(as.list(par), as.list(dtalist$fixed))

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    d_df[, new_var] <- with(args, with(d_df, eval(parse(text = formula))))
  }

  Xb <- d_df$Xb
  if ("theta_est" %in% colnames(d_df)) Xb <- Xb * d_df$theta_est

  logistic_tmp <- .5 + .5 * tanh(Xb / 2)
  pvals <- log(d_df$internal_lb *      logistic_tmp +
               (1 - d_df$internal_lb) * (1 - logistic_tmp))
  pvals <- matrix(pvals, nrow = nobs, ncol = 1)

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  pvals <- pvals[f_df$umatch, ]
  tout  <- -sum(pvals * f_df$internal_count)
  return(tout)
}

# ---------------------------------------------------------------------------
# Exported: continuous normal with random intercept (Gauss-Hermite quadrature)
# ---------------------------------------------------------------------------

#' Continuous normal log-likelihood with random intercept
#'
#' @title Continuous Normal Negative Log-Likelihood with Random Intercept
#' @description Computes the negative log-likelihood for a mixed-effects normal
#'   regression model with a random intercept, integrated via Gauss-Hermite
#'   quadrature. Designed as a \code{p_fun} argument for
#'   \code{\link{xreg2_control}}.
#'
#' @param par Named numeric vector of current parameter values. Must include
#'   either \code{LN_OMEGA} or \code{OMEGA} (between-subject SD).
#' @param dtalist Internal data list produced by \code{xreg2_fit}. Must
#'   include an \code{internal_id} column.
#' @param print_pars Logical; print parameters and value. Default \code{FALSE}.
#'
#' @return Scalar negative log-likelihood.
#'
#' @examples
#' # ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#' #                        p_fun = xreg2_cont_r_normal)
#'
#' @export
xreg2_cont_r_normal <- function(par, dtalist, print_pars = FALSE) {
  names(par) <- dtalist$coeff_names

  if (print_pars)
    message(do.call(paste, lapply(names(par),
                                  function(x) paste0(x, ": ", round(par[[x]], 3)))))

  formulas <- dtalist$formulas
  d_df <- dtalist$unique
  f_df <- dtalist$full

  args <- c(as.list(par), as.list(dtalist$fixed))

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    d_df[, new_var] <- with(args, with(d_df, eval(parse(text = formula))))
  }

  nobs <- NROW(d_df)

  if (!sum(c("LN_OMEGA", "OMEGA") %in% names(par)) == 1)
    stop("Ambiguous or non-unique OMEGA/LN_OMEGA parameter defined.")
  if ("LN_OMEGA" %in% names(par)) betwSD <- exp(par["LN_OMEGA"])
  if ("OMEGA"    %in% names(par)) betwSD <- par["OMEGA"]

  uid  <- sort(unique(f_df$internal_id))
  nuid <- NROW(uid)

  GH_order <- 8L
  if (!is.environment((xregenv <- getOption("xreg")))) {
    options(xreg = (xregenv <- new.env(parent = emptyenv())))
  }
  if (exists("Gauss_Hermite_Order", envir = xregenv)) {
    GH_order <- get("Gauss_Hermite_Order", envir = xregenv)
  } else {
    assign(x = "Gauss_Hermite_Order", value = GH_order, envir = xregenv)
  }
  ws   <- as.matrix(xreg_GHS[[GH_order]]$w)
  xs   <- xreg_GHS[[GH_order]]$x
  mus  <- sqrt(2) * betwSD * xs
  tmpm <- matrix(NA, nrow = nobs, ncol = GH_order)
  Xbm  <- Outer(y = d_df$Xb, x = mus, "+")

  sel  <- d_df$internal_type == 1
  sel2 <- !sel
  tmpm[sel, ]  <- dnorm(x    = d_df$internal_lb[sel],
                        mean = Xbm[sel, ],
                        sd   = d_df$sigma_est[sel],
                        log  = TRUE)
  tmpm[sel2, ] <- Log(
    pnorm(q = d_df$internal_lb[sel2], mean = Xbm[sel2, ],
          sd = d_df$sigma_est[sel2], lower.tail = FALSE) -
    pnorm(q = d_df$internal_ub[sel2], mean = Xbm[sel2, ],
          sd = d_df$sigma_est[sel2], lower.tail = FALSE)
  )

  tmpm <- tmpm[f_df$umatch, ] * f_df$internal_count

  pvals <- Log(1 / sqrt(pi) *
                 exp(apply(X      = tmpm,
                           MARGIN = 2,
                           FUN    = group.sum,
                           ina    = f_df$internal_id) +
                       rep_row(Log(ws), nuid)) %*% rep(1, GH_order))

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  tout <- -sum(pvals)
  if (is.na(tout)) {
    if (print_pars) message("Encountered NA")
    tout <- 10e10
  }
  if (print_pars) message(paste0("Value: ", tout))
  return(tout)
}

attr(x = xreg2_cont_r_normal, which = "required") <- list(
  sigma_est = list(
    formula      = sigma_est ~ exp(LN_SIGMA),
    start_values = c(LN_SIGMA = 0)
  ),
  omega_est = list(
    formula      = omega_est ~ exp(LN_OMEGA),
    start_values = c(OMEGA = 0)
  )
)

# ---------------------------------------------------------------------------
# Exported: continuous log-normal with random intercept
# ---------------------------------------------------------------------------

#' Continuous log-normal log-likelihood with random intercept
#'
#' @title Continuous Log-Normal Negative Log-Likelihood with Random Intercept
#' @description Computes the negative log-likelihood for a mixed-effects
#'   log-normal regression model with a random intercept, integrated via
#'   Gauss-Hermite quadrature. Designed as a \code{p_fun} argument for
#'   \code{\link{xreg2_control}}.
#'
#' @param par Named numeric vector of current parameter values. Must include
#'   either \code{LN_OMEGA} or \code{OMEGA}.
#' @param dtalist Internal data list produced by \code{xreg2_fit}. Must
#'   include an \code{internal_id} column.
#'
#' @return Scalar negative log-likelihood.
#'
#' @examples
#' # ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#' #                        p_fun = xreg2_cont_r_lnorm)
#'
#' @export
xreg2_cont_r_lnorm <- function(par, dtalist) {
  names(par) <- dtalist$coeff_names

  formulas <- dtalist$formulas
  d_df <- dtalist$unique
  f_df <- dtalist$full

  args <- c(as.list(par), as.list(dtalist$fixed))

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    d_df[, new_var] <- with(args, with(d_df, eval(parse(text = formula))))
  }

  nobs <- NROW(d_df)

  if (!sum(c("LN_OMEGA", "OMEGA") %in% names(par)) == 1)
    stop("Ambiguous or non-unique OMEGA/LN_OMEGA parameter defined.")
  if ("LN_OMEGA" %in% names(par)) betwSD <- exp(par["LN_OMEGA"])
  if ("OMEGA"    %in% names(par)) betwSD <- par["OMEGA"]

  uid  <- sort(unique(f_df$internal_id))
  nuid <- NROW(uid)

  GH_order <- 8L
  if (!is.environment((xregenv <- getOption("xreg")))) {
    options(xreg = (xregenv <- new.env(parent = emptyenv())))
  }
  if (exists("Gauss_Hermite_Order", envir = xregenv)) {
    GH_order <- get("Gauss_Hermite_Order", envir = xregenv)
  } else {
    assign(x = "Gauss_Hermite_Order", value = GH_order, envir = xregenv)
  }
  ws   <- as.matrix(xreg_GHS[[GH_order]]$w)
  xs   <- xreg_GHS[[GH_order]]$x
  mus  <- sqrt(2) * betwSD * xs
  tmpm <- matrix(NA, nrow = nobs, ncol = GH_order)
  Xbm  <- Outer(y = d_df$Xb, x = mus, "+")

  sel  <- d_df$internal_type == 1
  sel2 <- !sel
  tmpm[sel, ]  <- dlnorm(x       = d_df$internal_lb[sel],
                         meanlog = Xbm[sel, ],
                         sdlog   = d_df$sigma_est[sel],
                         log     = TRUE)
  tmpm[sel2, ] <- Log(
    plnorm(q = d_df$internal_ub[sel2], meanlog = Xbm[sel2, ],
           sdlog = d_df$sigma_est[sel2]) -
    plnorm(q = d_df$internal_lb[sel2], meanlog = Xbm[sel2, ],
           sdlog = d_df$sigma_est[sel2])
  )

  tmpm <- tmpm[f_df$umatch, ] * f_df$internal_count

  pvals <- Log(1 / sqrt(pi) *
                 exp(apply(X      = tmpm,
                           MARGIN = 2,
                           FUN    = group.sum,
                           ina    = f_df$internal_id) +
                       rep_row(Log(ws), nuid)) %*% rep(1, GH_order))

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  tout <- -sum(pvals)
  return(tout)
}

# ---------------------------------------------------------------------------
# Internal: gradient for cont_r_normal (3D array version)
# ---------------------------------------------------------------------------

grad_cont_r_normal1 <- function(par, dtalist, ret_mat = FALSE, eps = 1e-7, centered = TRUE) {

  if (centered) par <- par - eps / 2

  d_df <- dtalist$unique
  f_df <- dtalist$full
  nobs <- NROW(d_df)

  names(par) <- dtalist$coeff_names
  formulas   <- dtalist$formulas
  plen  <- length(par)
  plen1 <- plen + 1

  parm <- as.data.frame(as.list(par))
  parm <- parm[rep(1, plen + 1), ]
  for (pn in 1:plen) parm[pn + 1, pn] <- parm[pn + 1, pn] + eps

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    tmp <- sapply(1:plen1, function(x) {
      theseargs <- c(as.list(parm[x, ]), as.list(dtalist$fixed))
      with(theseargs, with(d_df, eval(parse(text = formula))))
    })
    if (is.null(dim(tmp)) && length(tmp) == plen1)
      tmp <- matrix(tmp, nrow = nobs, ncol = plen1, byrow = TRUE)
    d_df[, new_var] <- tmp
  }

  betwSD <- exp(parm[["LN_OMEGA"]])
  uid    <- sort(unique(f_df$internal_id))
  nuid   <- NROW(uid)

  GH_order <- 8L
  if (!is.environment((xregenv <- getOption("xreg")))) {
    options(xreg = (xregenv <- new.env(parent = emptyenv())))
  }
  if (exists("Gauss_Hermite_Order", envir = xregenv)) {
    GH_order <- get("Gauss_Hermite_Order", envir = xregenv)
  } else {
    assign(x = "Gauss_Hermite_Order", value = GH_order, envir = xregenv)
  }
  ws   <- as.matrix(xreg_GHS[[GH_order]]$w)
  xs   <- xreg_GHS[[GH_order]]$x
  mus  <- sqrt(2) * Outer(betwSD, xs)

  Xba  <- array(Outer(rep(0, GH_order), d_df$Xb, "+"), c(dim(d_df$Xb), GH_order))
  musa <- aperm(array(Outer(rep(0, nobs), mus, "+"), c(dim(mus), nobs)))
  pa   <- Xba <- Xba + musa

  sel  <- d_df$internal_type == 1
  sel2 <- !sel
  pa[sel, , ]  <- dnorm(x    = d_df$internal_lb[sel],
                        mean = Xba[sel, , ],
                        sd   = d_df$sigma_est[sel, ],
                        log  = TRUE)
  pa[sel2, , ] <- log(
    pnorm(q = d_df$internal_lb[sel2], mean = Xba[sel2, , ],
          sd = d_df$sigma_est[sel2, ], lower.tail = FALSE) -
    pnorm(q = d_df$internal_ub[sel2], mean = Xba[sel2, , ],
          sd = d_df$sigma_est[sel2, ], lower.tail = FALSE)
  )

  pa <- pa[f_df$umatch, , ]

  pvals <- apply(X = pa, MARGIN = 2, FUN = function(slice) {
    Log(1 / sqrt(pi) *
          exp(apply(X      = slice,
                    MARGIN = 2,
                    FUN    = group.sum,
                    ina    = f_df$internal_id) +
                rep_row(Log(ws), nuid)) %*% rep(1, GH_order))
  })

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  if (ret_mat) return(pvals)
  pvals <- colSums(pvals)
  pvals <- (pvals[1] - pvals[2:plen1]) / eps
  return(pvals)
}

# ---------------------------------------------------------------------------
# Internal: gradient for cont_r_normal (column-wise version, faster)
# ---------------------------------------------------------------------------

grad_cont_r_normal2 <- function(par, dtalist, ret_mat = FALSE, eps = 1e-7, centered = TRUE) {

  if (centered) par <- par - eps / 2

  d_df <- dtalist$unique
  f_df <- dtalist$full
  nobs <- NROW(d_df)

  names(par) <- dtalist$coeff_names
  formulas   <- dtalist$formulas
  plen  <- length(par)
  plen1 <- plen + 1

  parm <- as.data.frame(as.list(par))
  parm <- parm[rep(1, plen + 1), ]
  for (pn in 1:plen) parm[pn + 1, pn] <- parm[pn + 1, pn] + eps

  for (formulanum in seq_along(formulas)) {
    formula <- formulas[[formulanum]]
    new_var <- as.character(formula[[2]])
    if (new_var %in% c(colnames(d_df), dtalist$valuevar)) new_var <- "Xb"
    tmp <- sapply(1:plen1, function(x) {
      theseargs <- c(as.list(parm[x, ]), as.list(dtalist$fixed))
      with(theseargs, with(d_df, eval(parse(text = formula))))
    })
    if (is.null(dim(tmp)) && length(tmp) == plen1)
      tmp <- matrix(tmp, nrow = nobs, ncol = plen1, byrow = TRUE)
    d_df[, new_var] <- tmp
  }

  uid  <- sort(unique(f_df$internal_id))
  nuid <- NROW(uid)

  GH_order <- 8L
  if (!is.environment((xregenv <- getOption("xreg")))) {
    options(xreg = (xregenv <- new.env(parent = emptyenv())))
  }
  if (exists("Gauss_Hermite_Order", envir = xregenv)) {
    GH_order <- get("Gauss_Hermite_Order", envir = xregenv)
  } else {
    assign(x = "Gauss_Hermite_Order", value = GH_order, envir = xregenv)
  }
  ws <- as.matrix(xreg_GHS[[GH_order]]$w)
  xs <- xreg_GHS[[GH_order]]$x

  pvals <- sapply(1:plen1, function(x) {
    Xb     <- d_df$Xb[, x]
    betwSD <- exp(parm["LN_OMEGA"][x, ])
    mus    <- sqrt(2) * betwSD * xs
    tmpm   <- matrix(NA, nrow = nobs, ncol = GH_order)
    Xbm    <- Outer(y = Xb, x = mus, "+")

    sel  <- d_df$internal_type == 1
    sel2 <- !sel
    tmpm[sel, ]  <- dnorm(x    = d_df$internal_lb[sel],
                          mean = Xbm[sel, ],
                          sd   = d_df$sigma_est[sel, x],
                          log  = TRUE)
    tmpm[sel2, ] <- Log(
      pnorm(q = d_df$internal_lb[sel2], mean = Xbm[sel2, ],
            sd = d_df$sigma_est[sel2, x], lower.tail = FALSE) -
      pnorm(q = d_df$internal_ub[sel2], mean = Xbm[sel2, ],
            sd = d_df$sigma_est[sel2, x], lower.tail = FALSE)
    )

    tmpm <- tmpm[f_df$umatch, ] * f_df$internal_count

    Log(1 / sqrt(pi) *
          exp(apply(X      = tmpm,
                    MARGIN = 2,
                    FUN    = group.sum,
                    ina    = f_df$internal_id) +
                rep_row(Log(ws), nuid)) %*% rep(1, GH_order))
  })

  if ("boundNA" %in% names(dtalist))
    if (dtalist$boundNA) pvals[is.na(pvals)] <- log(.Machine$double.xmin)
  pvals[pvals == -Inf] <- log(.Machine$double.xmin)

  if (ret_mat) return(pvals)
  pvals <- colSums(pvals)
  pvals <- (pvals[1] - pvals[2:plen1]) / eps
  return(pvals)
}

# Internal: default gradient using numDeriv
grad_default <- function(func, x, method = "simple", ...) {
  numDeriv::grad(func = func, x = x, method = method, ...)
}

#' @title xreg2: Flexible Maximum Likelihood Regression with Gradient-Based Optimisation
#'
#' @description Fits flexible maximum likelihood regression models supporting
#'   censored, interval, and hybrid continuous/dichotomous data. Provides
#'   explicit analytic and numerical gradient computation, random intercept
#'   models via Gauss-Hermite quadrature, and multiple distribution families.
#'
#' @keywords internal
"_PACKAGE"

#' @importFrom stats AIC BIC aggregate as.formula complete.cases cor
#'   dbeta dexp dgamma dlnorm dnorm dweibull formula integrate na.omit
#'   pbeta pexp pgamma plnorm pnorm pt pweibull qnorm quantile sd var
NULL

# Statistical metric functions

#' Quantile summary statistics
#'
#' @title Quantile Summary
#' @description Computes summary statistics and quantiles for a numeric vector.
#'
#' @param x Numeric vector.
#' @param statfuns Function returning named summary statistics.
#' @param probs Named numeric vector of quantile probabilities.
#' @param names Logical; include names in quantile output.
#'
#' @return Named numeric vector of statistics and quantiles.
#' @export
quant_fun <- function(x,
                      statfuns = function(y) c(MEAN = mean(y, na.rm = TRUE),
                                               SE   = sd(y,   na.rm = TRUE)),
                      probs    = c("min" = 0, "2.5%" = 0.025, "25%" = 0.25,
                                   median = 0.5, "75%" = 0.75, "97.5%" = 0.975,
                                   "max"  = 1),
                      names    = TRUE) {
  c(statfuns(x), quantile(x, probs = probs, na.rm = TRUE, names = names))
}

#' Quantile summaries over matrix margins
#'
#' @title Quantile Summaries (Matrix)
#' @description Applies \code{\link{quant_fun}} over rows or columns of a
#'   matrix or data frame.
#'
#' @param x Matrix or data frame.
#' @param MARGIN Integer; \code{1} for rows, \code{2} for columns.
#' @param statfuns Function returning named summary statistics.
#' @param probs Named numeric vector of quantile probabilities.
#'
#' @return Data frame of statistics.
#' @export
quant_funs <- function(x,
                       MARGIN   = 1,
                       statfuns = function(y) c(MEAN = mean(y, na.rm = TRUE),
                                                SE   = sd(y,   na.rm = TRUE)),
                       probs    = c("min" = 0, "2.5%" = 0.025, "25%" = 0.25,
                                    median = 0.5, "75%" = 0.75, "97.5%" = 0.975,
                                    "max"  = 1)) {
  if (is.null(names(probs)))
    if (length(probs) > 0)
      if (all(probs <= 1) && all(probs >= 0))
        names(probs) <- paste0(round(probs * 100, 3), "%")
  as.data.frame(structure(
    .Data      = apply(X = x, MARGIN = MARGIN, quant_fun,
                       statfuns = statfuns, probs = probs, names = FALSE),
    .Dimnames  = list(
      c(names(as.list(as.list(statfuns)[[2]]))[-1], names(probs)),
      rownames(x)
    )
  ))
}

#' Transposed quantile summaries over matrix margins
#'
#' @title Transposed Quantile Summaries (Matrix)
#' @description Like \code{\link{quant_funs}} but returns transposed result.
#'
#' @param x Matrix or data frame.
#' @param MARGIN Integer; \code{1} for rows, \code{2} for columns.
#' @param statfuns Function returning named summary statistics.
#' @param probs Named numeric vector of quantile probabilities.
#'
#' @return Transposed data frame of statistics.
#' @export
quant_funst <- function(x,
                        MARGIN   = 1,
                        statfuns = function(y) c(MEAN = mean(y, na.rm = TRUE),
                                                 SE   = sd(y,   na.rm = TRUE)),
                        probs    = c("min" = 0, "2.5%" = 0.025, "25%" = 0.25,
                                     median = 0.5, "75%" = 0.75, "97.5%" = 0.975,
                                     "max"  = 1)) {
  as.data.frame(structure(
    .Data     = t(apply(X = x, MARGIN = MARGIN, quant_fun,
                        statfuns = statfuns, probs = probs, names = FALSE)),
    .Dimnames = list(
      rownames(x),
      c(names(as.list(as.list(statfuns)[[2]]))[-1], names(probs))
    )
  ))
}

#' Test for whole number
#'
#' @title Test for Whole Number
#' @description Tests whether numeric values are whole numbers within a
#'   numeric tolerance.
#'
#' @param x Numeric vector.
#' @param tol Numeric tolerance. Default \code{.Machine$double.eps^0.5}.
#'
#' @return Logical vector.
#' @export
is.wholenumber <- function(x, tol = .Machine$double.eps^0.5) abs(x - round(x)) < tol

#' Quantile summaries by group
#'
#' @title Quantile Summaries by Group
#' @description Computes quantile summaries within each level of a categorical
#'   variable.
#'
#' @param x Numeric vector.
#' @param catvar Factor or character vector of group labels.
#' @param pooled Logical; include a pooled row. Default \code{TRUE}.
#' @param n Logical; include count and missing count columns. Default \code{TRUE}.
#'
#' @return Data frame with one row per group.
#' @export
quant_fun_groups <- function(x, catvar, pooled = TRUE, n = TRUE) {
  if (inherits(catvar, "factor")) catvar <- levels(catvar)[as.integer(catvar)]
  tmp <- as.data.frame(t(mapply(function(gr) {
    y <- x[catvar %in% gr]
    if (gr == "Pooled") y <- x
    retv <- quant_fun(y)
    if (n) retv <- c(n = length(y), "n missing" = sum(is.na(y)), retv)
    return(retv)
  }, gr = c(unique(catvar), if (pooled) "Pooled" else NULL))))
  cbind(name = rownames(tmp), tmp)
}

#' @rdname quant_fun_groups
#' @param x A list of numeric vectors.
#' @export
quant_funs_groups <- function(x, catvar, pooled = TRUE, n = TRUE) {
  tmp <- lapply(x, FUN = quant_fun_groups, catvar = catvar, pooled = pooled, n = n)
  tmp <- Map(function(lv, ln) {
    lv <- rbind(NA, NA, lv)
    lv[2, 1] <- ln
    rownames(lv)[2] <- ln
    lv
  }, lv = tmp, ln = names(tmp))
  do.call(rbind, tmp)
}

#' Standardise distribution name
#'
#' @title Standardise Distribution Name
#' @description Maps common distribution name variants to their canonical R
#'   short form (e.g., \code{"normal"} -> \code{"norm"}).
#'
#' @param dist Character distribution name.
#'
#' @return Canonical short-form character name.
#' @export
fixdist <- function(dist) {
  structure(
    .Data  = c("norm", "norm", "lnorm", "lnorm", "lnorm", "exp", "exp",
               "beta", "gamma", "weibull", "weibull", tolower(dist)),
    .Names = c("normal", "norm", "lnorm", "lnormal", "lognormal", "exp",
               "exponential", "beta", "gamma", "weib", "weibull", tolower(dist))
  )[[tolower(dist)]]
}

#' Get a distribution function by type
#'
#' @title Get Distribution Function
#' @description Returns the named R distribution function (density, CDF,
#'   quantile, or random) for a given distribution, with an extra \code{...}
#'   argument appended to its formals.
#'
#' @param dist Character distribution name (passed to \code{\link{fixdist}}).
#' @param type Character; one of \code{"p"}, \code{"d"}, \code{"q"}, \code{"r"}.
#'
#' @return A function.
#' @export
getf <- function(dist, type) {
  thisf <- get(paste0(type, fixdist(dist)))
  formals(thisf) <- c(formals(thisf), alist("..." = ))
  return(thisf)
}

#' @rdname getf
#' @export
getp <- function(dist) getf(dist, "p")

#' @rdname getf
#' @export
getq <- function(dist) getf(dist, "q")

#' @rdname getf
#' @export
getd <- function(dist) getf(dist, "d")

#' @rdname getf
#' @export
getr <- function(dist) getf(dist, "r")

#' Mean of a distribution
#'
#' @param qfun Quantile function.
#' @param pars Named list of distribution parameters.
#' @return Numeric mean.
#' @export
getMean <- function(qfun, pars) do.call(integrate, c(list(qfun, 0, 1), pars))$value

#' Variance of a distribution
#'
#' @param qfun Quantile function or character distribution name.
#' @param pars Named list of distribution parameters.
#' @return Numeric variance.
#' @export
getVar <- function(qfun, pars) {
  if (is.character(qfun)) qfun <- getq(qfun)
  tmpmean <- getMean(qfun, pars)
  do.call(integrate,
          list(function(p) (do.call(qfun, c(list(p), pars)) - tmpmean)^2, 0, 1))$value
}

#' Standard deviation of a distribution
#'
#' @param qfun Quantile function or character distribution name.
#' @param pars Named list of distribution parameters.
#' @return Numeric standard deviation.
#' @export
getSD <- function(qfun, pars) sqrt(getVar(qfun, pars))

#' Density for an order statistic
#'
#' @title Order Statistic Density
#' @description Computes the density of the \eqn{r}-th order statistic in a
#'   random sample of \eqn{n} from a given distribution.
#'
#' @param x Numeric vector of values.
#' @param r Integer rank.
#' @param n Sample size.
#' @param par Named vector of distribution parameters.
#' @param dist Character distribution name.
#' @param log Logical; return log-density. Default \code{FALSE}.
#'
#' @return Numeric density (or log-density).
#' @export
ordD <- function(x, r, n, par = c(mean = 0, sd = 1), dist = "norm", log = FALSE) {
  if (any(r > n)) stop("r cannot be larger than n.")
  nchoices <- lfactorial(n) - (lfactorial(r - 1) + lfactorial(n - r))
  fx   <- do.call(getd(dist), c(list(x = x, log = TRUE), as.list(par)))
  Fxl  <- do.call(getp(dist), c(list(q = x, log = TRUE), as.list(par)))
  Fxlm <- do.call(getp(dist), c(list(q = x, log = TRUE, lower.tail = FALSE), as.list(par)))
  fx  [fx   == -Inf] <- log(.Machine$double.xmin)
  Fxl [Fxl  == -Inf] <- log(.Machine$double.xmin)
  Fxlm[Fxlm == -Inf] <- log(.Machine$double.xmin)
  if (log) {
    return(nchoices + fx + Fxlm * (n - r) + Fxl * (r - 1))
  } else {
    return(exp(nchoices + fx + Fxlm * (n - r) + Fxl * (r - 1)))
  }
}

#' Expected value of an order statistic
#'
#' @title Order Statistic Expected Value
#' @description Computes the expected value of the \eqn{r}-th order statistic
#'   in a random sample of \eqn{n} from a given distribution.
#'
#' @param r Integer rank (or vector of ranks).
#' @param n Sample size.
#' @param par Named vector of distribution parameters.
#' @param dist Character distribution name.
#' @param log Currently unused.
#'
#' @return Numeric expected value.
#' @export
ordE <- function(r, n, par = c(mean = 0, sd = 1), dist = "norm", log = FALSE) {
  if (any(r > n)) stop("r cannot be larger than n.")
  if (any(r < 1)) stop("r cannot be smaller than 1.")
  if (length(r) > 1) return(sapply(r, ordE, n = n, par = par, dist = dist, log = log))
  if (n < 1000) {
    bounds <- do.call(getq(dist), c(list(p = c(0.000001, 0.999999)), as.list(par)))
    integrate(f = function(x, ...) x * ordD(x, ...), lower = bounds[1], upper = bounds[2],
              r = r, n = n, par = par, dist = dist, log = log, subdivisions = 1000L)[[1]]
  } else {
    bounds <- do.call(getq(dist), c(list(p = c(0.000001, (1:9) / 10, 0.999999)), as.list(par)))
    sum(sapply(1:10, function(thisv) {
      integrate(f = function(x, ...) x * ordD(x, ...), lower = bounds[thisv],
                upper = bounds[thisv + 1], r = r, n = n, par = par,
                dist = dist, log = log, subdivisions = 1000L)[[1]]
    }))
  }
}

#' Expected probability of order statistics
#'
#' @title Order Statistic Expected Probabilities
#' @param probs Numeric vector of rank proportions.
#' @param n Sample size.
#' @param par Named vector of distribution parameters.
#' @param dist Character distribution name.
#' @return Numeric vector of probabilities.
#' @export
ordEp <- function(probs, n, par = c(mean = 0, sd = 1), dist = "normal") {
  r    <- n * probs
  names(r) <- probs
  pfun <- switch(dist,
    norm       = pnorm,
    normal     = pnorm,
    lnorm      = plnorm,
    lognormal  = plnorm,
    beta       = pbeta,
    gamma      = pgamma,
    exp        = pexp,
    exponential = pexp,
    weibull    = pweibull
  )
  newprobs <- sapply(r, function(x) {
    do.call(pfun, c(as.list(par), list(q = ordE(r = x, n = n, par = par, dist = dist))))
  })
  newprobs[(probs - newprobs) > 0.2] <- probs[(probs - newprobs) > 0.2]
  newprobs
}

# ---------------------------------------------------------------------------
# Fisher's ICC
# ---------------------------------------------------------------------------

#' Fisher's intraclass correlation coefficient
#'
#' @title Fisher's ICC
#' @description Computes Fisher's intraclass correlation coefficient between
#'   two numeric vectors or matrices.
#'
#' @param y Numeric vector, matrix, or data frame.
#' @param x Numeric vector, matrix, or data frame.
#' @param ... Currently unused.
#'
#' @return Numeric ICC value.
#' @export
fisher_icc <- function(y, x, ...) UseMethod("fisher_icc")

#' @rdname fisher_icc
#' @export
fisher_icc.data.frame <- function(y, x, ...) {
  apply(y, MARGIN = 2, FUN = fisher_icc, x, ...)
}

#' @rdname fisher_icc
#' @export
fisher_icc.matrix <- function(y, x, ...) fisher_icc(as.data.frame(y), x, ...)

#' @rdname fisher_icc
#' @export
fisher_icc.numeric <- function(y, x, ...) {
  if (!is.matrix(x) && is.numeric(x)) return(fisher_icc.two_numeric(y, x, ...))
  if (inherits(x, c("data.frame", "matrix")))
    return(apply(as.data.frame(x), MARGIN = 2, FUN = function(iter) fisher_icc(y, iter, ...)))
}

#' @exportS3Method
fisher_icc.two_numeric <- function(y, x, ...) {
  if (!is.numeric(y)) stop("y must be a numeric vector")
  if (!is.numeric(x)) stop("x must be a numeric vector")
  if (length(y) != length(x)) stop("y and x must have the same length")
  if (length(y) < 2) stop("y must be at least of length 2")
  if (all(y == x)) return(1)
  data    <- cbind(x, y)
  data    <- cbind(data, rowMeans(data))
  data    <- data[!is.na(data[, 3]), ]
  if (length(dim(data)) == 0) return(NA)
  x_avg   <- sum(data[, 1:2]) / (2 * nrow(data))
  data    <- cbind(data, (data[, 1:2] - x_avg)^2)
  s_sq    <- sum(data[, 4:5]) / (2 * (nrow(data) - 1))
  return(sum((data[, 1] - x_avg) * (data[, 2] - x_avg)) / ((nrow(data) - 1) * s_sq))
}

# ---------------------------------------------------------------------------
# Lin's concordance correlation coefficient
# ---------------------------------------------------------------------------

#' Lin's concordance correlation coefficient (simplified)
#'
#' @title Lin's CCC (simplified)
#' @param y Numeric vector, matrix, or data frame.
#' @param x Numeric vector, matrix, or data frame.
#' @param ... Passed to \code{\link{LCCC}}.
#' @return Numeric CCC value.
#' @export
just_CCC <- function(y, x, ...) UseMethod("just_CCC")

#' @rdname just_CCC
#' @export
just_CCC.data.frame <- function(y, x, ...) {
  apply(y, MARGIN = 2, FUN = just_CCC, x, ...)
}

#' @rdname just_CCC
#' @export
just_CCC.matrix <- function(y, x, ...) just_CCC(as.data.frame(y), x, ...)

#' @rdname just_CCC
#' @export
just_CCC.numeric <- function(y, x, ...) {
  if (!is.matrix(x) && is.numeric(x)) return(just_CCC.two_numeric(y, x, ...))
  if (inherits(x, c("data.frame", "matrix")))
    return(apply(as.data.frame(x), MARGIN = 2, FUN = function(iter) just_CCC(y, iter, ...)))
}

#' @exportS3Method
just_CCC.two_numeric <- function(y, x, ...) {
  if (!is.numeric(y)) stop("y must be a numeric vector")
  if (!is.numeric(x)) stop("x must be a numeric vector")
  if (length(y) != length(x)) stop("y and x must have the same length")
  if (length(y) < 2) return(NA)
  if (all(y == x)) return(1)
  LCCC(x, y, justp = TRUE, ...)
}

#' Lin's concordance correlation coefficient
#'
#' @title Lin's Concordance Correlation Coefficient
#' @description Calculates Lin's concordance correlation coefficient (CCC) for
#'   a pair of variables, measuring agreement along the 45-degree line.
#'
#' @param x Numeric vector; first variable.
#' @param y Numeric vector; second variable.
#' @param ci Character; method for confidence interval. Default
#'   \code{"z-transform"}.
#' @param conf.level Numeric; confidence level. Default \code{0.95}.
#' @param na.rm Logical; remove \code{NA} values. Default \code{FALSE}.
#' @param justp Logical; if \code{TRUE} return only the point estimate.
#'   Default \code{FALSE}.
#' @param ... Currently unused.
#'
#' @return A list with components \code{rho.c}, \code{s.shift},
#'   \code{l.shift}, \code{C.b}, and \code{blalt}.
#' @export
LCCC <- function(x, y, ci = "z-transform", conf.level = 0.95, na.rm = FALSE,
                 justp = FALSE, ...) {
  dat <- data.frame(x, y)
  dat <- if (na.rm) na.omit(dat) else dat
  N.  <- 1 - ((1 - conf.level) / 2)
  zv  <- qnorm(N.)
  k   <- length(dat$y)
  xb  <- mean(dat$x)
  yb  <- mean(dat$y)
  sx2 <- var(dat$x) * (k - 1) / k
  sy2 <- var(dat$y) * (k - 1) / k
  sd1 <- sd(dat$y)
  sd2 <- sd(dat$x)
  r   <- cor(dat$x, dat$y)
  sxy <- r * sqrt(sx2 * sy2)
  p   <- 2 * sxy / (sx2 + sy2 + (yb - xb)^2)
  if (justp) return(p)
  v   <- sd1 / sd2
  u   <- (yb - xb) / ((sx2 * sy2)^0.25)
  C.b <- p / r
  sep <- sqrt(((1 - r^2) * p^2 * (1 - p^2) / r^2 +
                 2 * p^3 * (1 - p) * u^2 / r -
                 0.5 * p^4 * u^4 / r^2) / (k - 2))
  ll  <- p - zv * sep
  ul  <- p + zv * sep
  t   <- log((1 + p) / (1 - p)) / 2
  set <- sep / (1 - p^2)
  llt <- t - zv * set
  ult <- t + zv * set
  llt <- (exp(2 * llt) - 1) / (exp(2 * llt) + 1)
  ult <- (exp(2 * ult) - 1) / (exp(2 * ult) + 1)
  rho.c <- data.frame(est    = p,
                      lwr.ci = ifelse(ci == "asymptotic", ll, llt),
                      upr.ci = ifelse(ci == "asymptotic", ul, ult))
  list(rho.c  = rho.c,
       s.shift = v,
       l.shift = u,
       C.b     = C.b,
       blalt   = data.frame(mean = rowMeans(dat), delta = dat$x - dat$y))
}

# ---------------------------------------------------------------------------
# Error metrics (S3 generics)
# ---------------------------------------------------------------------------

#' Average Relative Error
#'
#' @title Average Relative Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param absolute Logical; return absolute version. Default \code{FALSE}.
#' @param ... Currently unused.
#' @return Numeric ARE value.
#' @export
ARE <- function(est, obs, absolute = FALSE, ...) UseMethod("ARE")

#' @rdname ARE
#' @export
ARE.data.frame <- function(est, obs, absolute = FALSE, ...) {
  apply(est, MARGIN = 2, FUN = ARE, obs, absolute, ...)
}

#' @rdname ARE
#' @export
ARE.matrix <- function(est, obs, absolute = FALSE, ...) ARE(as.data.frame(est), obs, absolute, ...)

#' @rdname ARE
#' @export
ARE.numeric <- function(est, obs, absolute = FALSE, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(ARE.two_numeric(est, obs, absolute, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) ARE(est, iter, absolute, ...)))
}

#' @exportS3Method
ARE.two_numeric <- function(est, obs, absolute = FALSE, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  if (absolute) return(mean(abs(est - obs) / obs))
  return(mean((est - obs) / obs))
}

#' @rdname ARE
#' @param dfys Numeric matrix or data frame of estimates.
#' @param dfobs Numeric matrix or data frame of observations.
#' @export
ARE_conform <- function(dfys, dfobs, absolute = FALSE) {
  if (is.list(dfys)) return(sapply(dfys, function(dfest) ARE_conform(dfest, dfobs, absolute)))
  if (is.matrix(dfobs)) {
    if (is.null(colnames(dfobs)) && !is.null(rownames(dfobs))) {
      dfys  <- t(dfys)
      dfobs <- t(dfobs)
    }
  }
  dfest <- as.data.frame(dfys)
  dfobs <- as.data.frame(dfobs)
  nx    <- colnames(dfobs)
  colnames(dfest) <- nx
  names(nx) <- nx
  sapply(nx, function(thisn) ARE(dfest[, thisn], dfobs[, thisn], absolute))
}

#' Relative Mean Squared Error
#'
#' @title Relative Mean Squared Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param trueobs Numeric; true values.
#' @param ... Currently unused.
#' @return Numeric ReMSE value.
#' @export
ReMSE <- function(est, obs, trueobs, ...) UseMethod("ReMSE")

#' @rdname ReMSE
#' @export
ReMSE.data.frame <- function(est, obs, trueobs, ...) {
  apply(est, MARGIN = 2, FUN = ReMSE, obs, trueobs, ...)
}

#' @rdname ReMSE
#' @export
ReMSE.matrix <- function(est, obs, trueobs, ...) ReMSE(as.data.frame(est), obs, trueobs, ...)

#' @rdname ReMSE
#' @export
ReMSE.numeric <- function(est, obs, trueobs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(ReMSE.two_numeric(est, obs, trueobs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) ReMSE(est, iter, trueobs, ...)))
}

#' @exportS3Method
ReMSE.two_numeric <- function(est, obs, trueobs, ...) {
  if (!is.numeric(est))     stop("est must be a numeric vector")
  if (!is.numeric(obs))     stop("obs must be a numeric vector")
  if (!is.numeric(trueobs)) stop("trueobs must be a numeric vector")
  return(mean((est - trueobs)^2) / mean((obs - trueobs)^2))
}

#' @rdname ReMSE
#' @param dfys Numeric matrix or data frame of estimates.
#' @param dfobs Numeric matrix or data frame of observations.
#' @export
ReMSE_conform <- function(dfys, dfobs, trueobs) {
  if (is.list(dfys)) return(sapply(dfys, function(dfest) ReMSE_conform(dfest, dfobs, trueobs)))
  if (is.matrix(dfobs)) {
    if (is.null(colnames(dfobs)) && !is.null(rownames(dfobs))) {
      dfys  <- t(dfys)
      dfobs <- t(dfobs)
    }
  }
  dfest <- as.data.frame(dfys)
  dfobs <- as.data.frame(dfobs)
  nx    <- colnames(dfobs)
  colnames(dfest) <- nx
  names(nx) <- nx
  if (length(trueobs) == 1) trueobs <- rep(trueobs, length(nx))
  names(trueobs) <- nx
  trueobs <- as.list(trueobs)
  sapply(nx, function(thisn) ReMSE(dfest[, thisn], dfobs[, thisn], trueobs[[thisn]]))
}

#' Root Mean Square Error
#'
#' @title Root Mean Square Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param percentage Logical; return as percentage of mean. Default \code{FALSE}.
#' @param ... Passed to \code{mean} and \code{sqrt}.
#' @return Numeric RMSE value.
#' @export
RMSE <- function(est, obs, percentage = FALSE, ...) UseMethod("RMSE")

#' @rdname RMSE
#' @export
RMSE.data.frame <- function(est, obs, percentage = FALSE, ...) {
  apply(est, MARGIN = 2, FUN = RMSE, obs, percentage, ...)
}

#' @rdname RMSE
#' @export
RMSE.matrix <- function(est, obs, percentage = FALSE, ...) {
  RMSE(as.data.frame(est), obs, percentage, ...)
}

#' @rdname RMSE
#' @export
RMSE.numeric <- function(est, obs, percentage = FALSE, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(RMSE.two_numeric(est, obs, percentage, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) RMSE(est, iter, percentage, ...)))
}

#' @exportS3Method
RMSE.two_numeric <- function(est, obs, percentage = FALSE, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  if (percentage) {
    return(sqrt(mean((obs - est)^2), ...) / mean(obs, ...) * 100)
  } else {
    return(sqrt(mean((obs - est)^2, ...)))
  }
}

#' Percentage RMSE
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Passed to \code{\link{RMSE}}.
#' @rdname RMSE
#' @export
pRMSE <- function(est, obs, ...) RMSE(est, obs, percentage = TRUE, ...)

#' @rdname RMSE
#' @export
rRMSE <- function(est, obs, ...) RMSE(est, obs, percentage = TRUE, ...)

#' Mean Absolute Error
#'
#' @title Mean Absolute Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Passed to \code{mean}.
#' @return Numeric MAE value.
#' @export
MAE <- function(est, obs, ...) UseMethod("MAE")

#' @rdname MAE
#' @export
MAE.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = MAE, obs, ...)
}

#' @rdname MAE
#' @export
MAE.matrix <- function(est, obs, ...) MAE(as.data.frame(est), obs, ...)

#' @rdname MAE
#' @export
MAE.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(MAE.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) MAE(est, iter, ...)))
}

#' @exportS3Method
MAE.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(mean(abs(obs - est), ...))
}

#' Mean Absolute Relative Error / Mean Absolute Percentage Error
#'
#' @title Mean Absolute Relative Error (MARE / MAPE)
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Currently unused.
#' @return Numeric MARE value.
#' @export
MAPE <- function(est, obs, ...) UseMethod("MARE")

#' @rdname MAPE
#' @export
MARE <- function(est, obs, ...) UseMethod("MARE")

#' @rdname MAPE
#' @export
MARE.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = MARE, obs, ...)
}

#' @rdname MAPE
#' @export
MARE.matrix <- function(est, obs, ...) MARE(as.data.frame(est), obs, ...)

#' @rdname MAPE
#' @export
MARE.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(MARE.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) MARE(est, iter, ...)))
}

#' @exportS3Method
MARE.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(mean(abs((obs - est) / obs), ...))
}

#' Uncertainty at 95 Percent
#'
#' @title Uncertainty at 95 Percent
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Passed to \code{var} and \code{sqrt}.
#' @return Numeric U95 value.
#' @export
U95 <- function(est, obs, ...) UseMethod("U95")

#' @rdname U95
#' @export
U95.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = U95, obs, ...)
}

#' @rdname U95
#' @export
U95.matrix <- function(est, obs, ...) U95(as.data.frame(est), obs, ...)

#' @rdname U95
#' @export
U95.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(U95.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) U95(est, iter, ...)))
}

#' @exportS3Method
U95.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(1.96 * sqrt(var(obs - est, ...) + RMSE(est, obs, ...)^2, ...))
}

#' Root Mean Square Relative Error
#'
#' @title Root Mean Square Relative Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Currently unused.
#' @return Numeric RMSRE value.
#' @export
RMSRE <- function(est, obs, ...) UseMethod("RMSRE")

#' @rdname RMSRE
#' @export
RMSRE.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = RMSRE, obs, ...)
}

#' @rdname RMSRE
#' @export
RMSRE.matrix <- function(est, obs, ...) RMSRE(as.data.frame(est), obs, ...)

#' @rdname RMSRE
#' @export
RMSRE.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(RMSRE.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) RMSRE(est, iter, ...)))
}

#' @exportS3Method
RMSRE.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(sqrt(mean(((obs - est) / obs)^2, ...), ...))
}

#' Mean Bias Error
#'
#' @title Mean Bias Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Passed to \code{mean}.
#' @return Numeric MBE value.
#' @export
MBE <- function(est, obs, ...) UseMethod("MBE")

#' @rdname MBE
#' @export
MBE.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = MBE, obs, ...)
}

#' @rdname MBE
#' @export
MBE.matrix <- function(est, obs, ...) MBE(as.data.frame(est), obs, ...)

#' @rdname MBE
#' @export
MBE.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(MBE.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) MBE(est, iter, ...)))
}

#' @exportS3Method
MBE.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(mean(obs - est, ...))
}

#' Maximum Absolute Relative Error
#'
#' @title Maximum Absolute Relative Error
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Currently unused.
#' @return Numeric erMAX value.
#' @export
erMAX <- function(est, obs, ...) UseMethod("erMAX")

#' @rdname erMAX
#' @export
erMAX.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = erMAX, obs, ...)
}

#' @rdname erMAX
#' @export
erMAX.matrix <- function(est, obs, ...) erMAX(as.data.frame(est), obs, ...)

#' @rdname erMAX
#' @export
erMAX.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(erMAX.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) erMAX(est, iter, ...)))
}

#' @exportS3Method
erMAX.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  return(max(abs((obs - est) / obs, ...), ...))
}

#' t-statistic for model comparison
#'
#' @title t-Statistic
#' @param est Numeric; estimates.
#' @param obs Numeric; observations.
#' @param ... Currently unused.
#' @return Numeric t-statistic.
#' @export
tstat <- function(est, obs, ...) UseMethod("tstat")

#' @rdname tstat
#' @export
tstat.data.frame <- function(est, obs, ...) {
  apply(est, MARGIN = 2, FUN = tstat, obs, ...)
}

#' @rdname tstat
#' @export
tstat.matrix <- function(est, obs, ...) tstat(as.data.frame(est), obs, ...)

#' @rdname tstat
#' @export
tstat.numeric <- function(est, obs, ...) {
  if (!is.matrix(obs) && is.numeric(obs)) return(tstat.two_numeric(est, obs, ...))
  if (inherits(obs, c("data.frame", "matrix")))
    return(apply(as.data.frame(obs), MARGIN = 2,
                 FUN = function(iter) tstat(est, iter, ...)))
}

#' @exportS3Method
tstat.two_numeric <- function(est, obs, ...) {
  if (!is.numeric(est)) stop("est must be a numeric vector")
  if (!is.numeric(obs)) stop("obs must be a numeric vector")
  thisMBEs <- MBE(est, obs, ...)^2
  return(sqrt(((sum(!is.na(obs)) - 1) * thisMBEs) / (RMSE(est, obs, ...)^2 - thisMBEs)))
}

#' Comparison table of error metrics
#'
#' @title Comparison Table
#' @description Computes multiple error metrics and returns them as a matrix.
#'
#' @param est Numeric vector or matrix of estimates.
#' @param obs Numeric vector of observations.
#' @param ... Passed to metric functions.
#'
#' @return A numeric matrix with one row per metric.
#' @export
comp_table <- function(est, obs, ...) {
  rbind("Pearson's R" = cor(est, obs, ...)[, ],
        ICC    = fisher_icc(est, obs, ...),
        CCC    = just_CCC(est, obs, ...),
        MAE    = MAE(est, obs, ...),
        RMSE   = RMSE(est, obs, ...),
        MARE   = MARE(est, obs, ...),
        U95    = U95(est, obs, ...),
        RMSRE  = RMSRE(est, obs, ...),
        rRMSE  = rRMSE(est, obs, ...),
        erMAX  = erMAX(est, obs, ...),
        MBE    = MBE(est, obs, ...),
        tstat  = tstat(est, obs, ...))
}

#' Two-sample t-test from summary statistics
#'
#' @title Two-Sample t-Test from Summary Statistics
#' @param m1 Numeric; mean of sample 1.
#' @param m2 Numeric; mean of sample 2.
#' @param s1 Numeric; SD of sample 1.
#' @param s2 Numeric; SD of sample 2.
#' @param n1 Numeric; size of sample 1.
#' @param n2 Numeric; size of sample 2.
#' @param m0 Numeric; null difference. Default \code{0}.
#' @param equal.variance Logical; assume equal variances. Default \code{FALSE}.
#' @return Named numeric vector with test statistics and confidence interval.
#' @export
t_test2 <- function(m1, m2, s1, s2, n1, n2, m0 = 0, equal.variance = FALSE) {
  if (!equal.variance) {
    se <- sqrt((s1^2 / n1) + (s2^2 / n2))
    df <- ((s1^2 / n1 + s2^2 / n2)^2) / ((s1^2 / n1)^2 / (n1 - 1) + (s2^2 / n2)^2 / (n2 - 1))
  } else {
    se <- sqrt((1 / n1 + 1 / n2) * ((n1 - 1) * s1^2 + (n2 - 1) * s2^2) / (n1 + n2 - 2))
    df <- n1 + n2 - 2
  }
  t   <- (m1 - m2 - m0) / se
  md  <- m1 - m2
  dat <- c(md, se, t, 2 * pt(-abs(t), df), md + se * qnorm(c(0.025, 0.975)))
  names(dat) <- c("Difference of means", "Std Error", "t", "p-value", "95% LB", "95% UB")
  return(dat)
}

#' Combined mean of two groups
#' @param m1 Mean of group 1 (scalar or vector).
#' @param m2 Mean of group 2 (scalar or vector).
#' @param n1 Size of group 1.
#' @param n2 Size of group 2.
#' @param na.rm Logical.
#' @return Combined mean.
#' @export
combmean <- function(m1, m2, n1 = NULL, n2 = NULL, na.rm = FALSE) {
  if (any(is.na(m1))) { n1 <- n1[!is.na(m1)]; m1 <- m1[!is.na(m1)] }
  if (any(is.na(m2))) { n2 <- n2[!is.na(m2)]; m2 <- m2[!is.na(m2)] }
  if (length(m1) > 1 || length(m2) > 1)
    return(combmean(mean(m1, na.rm = na.rm), mean(m2, na.rm = na.rm),
                    length(m1), length(m2)))
  sum(n1 * m1, n2 * m2, na.rm = na.rm) / sum(n1, n2, na.rm = na.rm)
}

#' Combined standard deviation of two groups
#' @param m1 Mean of group 1.
#' @param m2 Mean of group 2.
#' @param s1 SD of group 1.
#' @param s2 SD of group 2.
#' @param n1 Size of group 1.
#' @param n2 Size of group 2.
#' @param na.rm Logical.
#' @return Combined SD.
#' @export
combsd <- function(m1, m2, s1, s2, n1, n2, na.rm = FALSE) {
  df1 <- data.frame(m = m1, s = s1, n = n1)
  df2 <- data.frame(m = m2, s = s2, n = n2)
  if (na.rm) { df1 <- df1[complete.cases(df1), ]; df2 <- df2[complete.cases(df2), ] }
  df <- rbind(df1, df2)
  if (NROW(df) == 1) {
    df$s
  } else {
    m12 <- combmean(m1, m2, n1, n2, na.rm = na.rm)
    d1  <- sum(m1, -m12, na.rm = na.rm)
    d2  <- sum(m2, -m12, na.rm = na.rm)
    sqrt((n1 * (s1^2 + d1^2) + n2 * (s2^2 + d2^2)) / (n1 + n2))
  }
}

#' Combined mean of multiple groups
#' @param ms Numeric vector of group means.
#' @param ns Numeric vector of group sizes.
#' @param na.rm Logical.
#' @return Combined mean.
#' @export
combmeans <- function(ms, ns, na.rm = FALSE) {
  cm <- 0
  cn <- 0
  for (i in seq_along(ms)) {
    cm <- combmean(cm, ms[i], cn, ns[i], na.rm = na.rm)
    cn <- cn + ns[i]
  }
  return(cm)
}

#' Combined SD of multiple groups
#' @param ms Numeric vector of group means.
#' @param ss Numeric vector of group SDs.
#' @param ns Numeric vector of group sizes.
#' @param na.rm Logical.
#' @return Combined SD.
#' @export
combsds <- function(ms, ss, ns, na.rm = FALSE) {
  cm <- 0
  cs <- 0
  cn <- 0
  for (i in seq_along(ms)) {
    cs <- combsd(cm, ms[i], cs, ss[i], cn, ns[i], na.rm = na.rm)
    cm <- combmean(cm, ms[i], cn, ns[i])
    cn <- cn + ns[i]
  }
  return(cs)
}

#' Combined mean, SD and n for multiple groups
#' @param ms Numeric vector of group means.
#' @param ss Numeric vector of group SDs.
#' @param ns Numeric vector of group sizes.
#' @param na.rm Logical.
#' @return Matrix with columns n, mean, sd.
#' @export
combmssds <- function(ms, ss, ns, na.rm = FALSE) {
  cm <- 0
  cs <- 0
  cn <- 0
  for (i in seq_along(ms)) {
    if (na.rm) {
      if (!any(is.na(c(ms[i], ss[i], ns[i])))) {
        cs <- combsd(cm, ms[i], cs, ss[i], cn, ns[i], na.rm = na.rm)
        cm <- combmean(cm, ms[i], cn, ns[i], na.rm = na.rm)
        cn <- sum(cn, ns[i], na.rm = na.rm)
      }
    } else {
      cs <- combsd(cm, ms[i], cs, ss[i], cn, ns[i], na.rm = na.rm)
      cm <- combmean(cm, ms[i], cn, ns[i], na.rm = na.rm)
      cn <- sum(cn, ns[i], na.rm = na.rm)
    }
  }
  return(cbind(n = cn, mean = cm, sd = cs))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{ordEp}
\alias{ordEp}
\title{Order Statistic Expected Probabilities}
\usage{
ordEp(probs, n, par = c(mean = 0, sd = 1), dist = "normal")
}
\arguments{
\item{probs}{Numeric vector of rank proportions.}

\item{n}{Sample size.}

\item{par}{Named vector of distribution parameters.}

\item{dist}{Character distribution name.}
}
\value{
Numeric vector of probabilities.
}
\description{
Expected probability of order statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ll_functions.R
\name{xreg2_dich_logistic}
\alias{xreg2_dich_logistic}
\title{Dichotomous Logistic Negative Log-Likelihood}
\usage{
xreg2_dich_logistic(par, dtalist)
}
\arguments{
\item{par}{Named numeric vector of current parameter values.}

\item{dtalist}{Internal data list produced by \code{xreg2_fit}.}
}
\value{
Scalar negative log-likelihood.
}
\description{
Computes the negative log-likelihood for a binary logistic
  regression model. Designed as a \code{p_fun} argument for
  \code{\link{xreg2_control}}.
}
\details{
Dichotomous logistic log-likelihood for xreg2
}
\examples{
# ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#                        p_fun = xreg2_dich_logistic,
#                        start_values = c(INTERCEPT = 0, SLOPE = 0))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{MBE}
\alias{MBE}
\alias{MBE.data.frame}
\alias{MBE.matrix}
\alias{MBE.numeric}
\title{Mean Bias Error}
\usage{
MBE(est, obs, ...)

\method{MBE}{data.frame}(est, obs, ...)

\method{MBE}{matrix}(est, obs, ...)

\method{MBE}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Passed to \code{mean}.}
}
\value{
Numeric MBE value.
}
\description{
Mean Bias Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{AIC.xreg2}
\alias{AIC.xreg2}
\title{AIC for an xreg2 Model}
\usage{
\method{AIC}{xreg2}(object, ..., k = 2, just_total = TRUE)
}
\arguments{
\item{object}{A fitted \code{xreg2} object.}

\item{...}{Currently unused.}

\item{k}{Numeric; penalty per parameter. Default \code{2}.}

\item{just_total}{Logical; return only the total AIC across all components.
Default \code{TRUE}.}
}
\value{
A numeric scalar (AIC value).
}
\description{
Computes the Akaike Information Criterion for a fitted
  \code{xreg2} model. Scale parameters (SIGMA, LN_SIGMA, OMEGA, LN_OMEGA)
  are excluded from the parameter count.
}
\details{
AIC method for xreg2 objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{ARE}
\alias{ARE}
\alias{ARE.data.frame}
\alias{ARE.matrix}
\alias{ARE.numeric}
\alias{ARE_conform}
\title{Average Relative Error}
\usage{
ARE(est, obs, absolute = FALSE, ...)

\method{ARE}{data.frame}(est, obs, absolute = FALSE, ...)

\method{ARE}{matrix}(est, obs, absolute = FALSE, ...)

\method{ARE}{numeric}(est, obs, absolute = FALSE, ...)

ARE_conform(dfys, dfobs, absolute = FALSE)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{absolute}{Logical; return absolute version. Default \code{FALSE}.}

\item{...}{Currently unused.}

\item{dfys}{Numeric matrix or data frame of estimates.}

\item{dfobs}{Numeric matrix or data frame of observations.}
}
\value{
Numeric ARE value.
}
\description{
Average Relative Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hyreg.R
\name{xreg2_hyreg}
\alias{xreg2_hyreg}
\title{Hybrid Regression (xreg2 version)}
\usage{
xreg2_hyreg(
  formula,
  df,
  datatype = "method",
  init = numeric(),
  contdist = "normal",
  dichdist = "logistic",
  hetcont = NULL,
  hetdich = NULL,
  ul = Inf,
  ll = -Inf,
  lntheta = TRUE,
  lnsigma = TRUE,
  dichformula = NULL,
  ...
)
}
\arguments{
\item{formula}{A formula or list of formulas for the continuous component.
The dependent variable (left-hand side) must match a column in \code{df},
or use interval columns named \code{varname.lb} and \code{varname.ub}.}

\item{df}{A \code{data.frame} containing both continuous and dichotomous
observations. Must include the column specified by \code{datatype}.}

\item{datatype}{Character; name of the column that distinguishes continuous
(\code{TRUE}) from dichotomous (\code{FALSE}) observations. Default
\code{"method"}.}

\item{init}{Optional named numeric vector (or \code{xreg2} object) of
starting values.}

\item{contdist}{Character; distribution for the continuous component.
\code{"normal"} (default) or \code{"logistic"}.}

\item{dichdist}{Character; distribution for the dichotomous component.
\code{"logistic"} (default) or \code{"normal"}.}

\item{hetcont}{Optional formula for heteroscedastic standard deviations in
the continuous component. If \code{NULL} (default), uses
\code{sigma_est ~ SIGMA} (or \code{sigma_est ~ exp(SIGMA)} when
\code{lnsigma = TRUE}).}

\item{hetdich}{Optional formula for heteroscedastic standard deviations in
the dichotomous component. If \code{NULL} (default), uses
\code{theta_est ~ THETA} (or \code{theta_est ~ exp(THETA)} when
\code{lntheta = TRUE}).}

\item{ul}{Numeric; upper censoring bound. Default \code{Inf}.}

\item{ll}{Numeric; lower censoring bound. Default \code{-Inf}.}

\item{lntheta}{Logical; exponentiate theta prior to fitting (matches STATA
default). Default \code{TRUE}.}

\item{lnsigma}{Logical; exponentiate sigma prior to fitting (matches STATA
default). Default \code{TRUE}.}

\item{dichformula}{Optional formula for the dichotomous component. Defaults
to \code{formula}.}

\item{...}{Additional arguments forwarded to \code{\link{xreg2_fit}}.}
}
\value{
An object of class \code{"xreg2"}.
}
\description{
High-level wrapper around \code{\link{xreg2_fit}} replicating
  the interface of the STATA \code{hyreg} command and
  \code{xreg::hyreg()}. Simultaneously fits a censored continuous model
  (normal or logistic) and a dichotomous model (logistic or normal) to a
  single data frame that contains both observation types.
}
\details{
Hybrid regression wrapper for xreg2
}
\examples{
\donttest{
set.seed(1)
n   <- 300
df  <- data.frame(
  value  = c(rnorm(200, 0.6, 0.3), rep(NA, 100)),
  age    = rnorm(n),
  method = c(rep(TRUE, 200), rep(FALSE, 100))
)
df$value[df$method == FALSE] <- as.integer(df$age[df$method == FALSE] > 0)

fit <- xreg2_hyreg(
  formula  = value ~ INTERCEPT + age * AGE,
  df       = df,
  datatype = "method"
)
print(fit)
}

}
\seealso{
\code{\link{xreg2_fit}}, \code{\link{xreg2_control}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{fixdist}
\alias{fixdist}
\title{Standardise Distribution Name}
\usage{
fixdist(dist)
}
\arguments{
\item{dist}{Character distribution name.}
}
\value{
Canonical short-form character name.
}
\description{
Maps common distribution name variants to their canonical R
  short form (e.g., \code{"normal"} -> \code{"norm"}).
}
\details{
Standardise distribution name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{RMSE}
\alias{RMSE}
\alias{RMSE.data.frame}
\alias{RMSE.matrix}
\alias{RMSE.numeric}
\alias{pRMSE}
\alias{rRMSE}
\title{Root Mean Square Error}
\usage{
RMSE(est, obs, percentage = FALSE, ...)

\method{RMSE}{data.frame}(est, obs, percentage = FALSE, ...)

\method{RMSE}{matrix}(est, obs, percentage = FALSE, ...)

\method{RMSE}{numeric}(est, obs, percentage = FALSE, ...)

pRMSE(est, obs, ...)

rRMSE(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{percentage}{Logical; return as percentage of mean. Default \code{FALSE}.}

\item{...}{Passed to \code{\link{RMSE}}.}
}
\value{
Numeric RMSE value.
}
\description{
Root Mean Square Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{RMSRE}
\alias{RMSRE}
\alias{RMSRE.data.frame}
\alias{RMSRE.matrix}
\alias{RMSRE.numeric}
\title{Root Mean Square Relative Error}
\usage{
RMSRE(est, obs, ...)

\method{RMSRE}{data.frame}(est, obs, ...)

\method{RMSRE}{matrix}(est, obs, ...)

\method{RMSRE}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Currently unused.}
}
\value{
Numeric RMSRE value.
}
\description{
Root Mean Square Relative Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{tstat}
\alias{tstat}
\alias{tstat.data.frame}
\alias{tstat.matrix}
\alias{tstat.numeric}
\title{t-Statistic}
\usage{
tstat(est, obs, ...)

\method{tstat}{data.frame}(est, obs, ...)

\method{tstat}{matrix}(est, obs, ...)

\method{tstat}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Currently unused.}
}
\value{
Numeric t-statistic.
}
\description{
t-statistic for model comparison
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{getVar}
\alias{getVar}
\title{Variance of a distribution}
\usage{
getVar(qfun, pars)
}
\arguments{
\item{qfun}{Quantile function or character distribution name.}

\item{pars}{Named list of distribution parameters.}
}
\value{
Numeric variance.
}
\description{
Variance of a distribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{LCCC}
\alias{LCCC}
\title{Lin's Concordance Correlation Coefficient}
\usage{
LCCC(
  x,
  y,
  ci = "z-transform",
  conf.level = 0.95,
  na.rm = FALSE,
  justp = FALSE,
  ...
)
}
\arguments{
\item{x}{Numeric vector; first variable.}

\item{y}{Numeric vector; second variable.}

\item{ci}{Character; method for confidence interval. Default
\code{"z-transform"}.}

\item{conf.level}{Numeric; confidence level. Default \code{0.95}.}

\item{na.rm}{Logical; remove \code{NA} values. Default \code{FALSE}.}

\item{justp}{Logical; if \code{TRUE} return only the point estimate.
Default \code{FALSE}.}

\item{...}{Currently unused.}
}
\value{
A list with components \code{rho.c}, \code{s.shift},
  \code{l.shift}, \code{C.b}, and \code{blalt}.
}
\description{
Calculates Lin's concordance correlation coefficient (CCC) for
  a pair of variables, measuring agreement along the 45-degree line.
}
\details{
Lin's concordance correlation coefficient
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{combmssds}
\alias{combmssds}
\title{Combined mean, SD and n for multiple groups}
\usage{
combmssds(ms, ss, ns, na.rm = FALSE)
}
\arguments{
\item{ms}{Numeric vector of group means.}

\item{ss}{Numeric vector of group SDs.}

\item{ns}{Numeric vector of group sizes.}

\item{na.rm}{Logical.}
}
\value{
Matrix with columns n, mean, sd.
}
\description{
Combined mean, SD and n for multiple groups
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{BIC.xreg2}
\alias{BIC.xreg2}
\title{BIC for an xreg2 Model}
\usage{
\method{BIC}{xreg2}(object, ..., just_total = TRUE)
}
\arguments{
\item{object}{A fitted \code{xreg2} object.}

\item{...}{Currently unused.}

\item{just_total}{Logical; return only the total BIC. Default \code{TRUE}.}
}
\value{
A numeric scalar (BIC value).
}
\description{
Computes the Bayesian Information Criterion for a fitted
  \code{xreg2} model.
}
\details{
BIC method for xreg2 objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{getf}
\alias{getf}
\alias{getp}
\alias{getq}
\alias{getd}
\alias{getr}
\title{Get Distribution Function}
\usage{
getf(dist, type)

getp(dist)

getq(dist)

getd(dist)

getr(dist)
}
\arguments{
\item{dist}{Character distribution name (passed to \code{\link{fixdist}}).}

\item{type}{Character; one of \code{"p"}, \code{"d"}, \code{"q"}, \code{"r"}.}
}
\value{
A function.
}
\description{
Returns the named R distribution function (density, CDF,
  quantile, or random) for a given distribution, with an extra \code{...}
  argument appended to its formals.
}
\details{
Get a distribution function by type
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{quant_fun}
\alias{quant_fun}
\title{Quantile Summary}
\usage{
quant_fun(
  x,
  statfuns = function(y) c(MEAN = mean(y, na.rm = TRUE), SE = sd(y, na.rm = TRUE)),
  probs = c(min = 0, `2.5\%` = 0.025, `25\%` = 0.25, median = 0.5, `75\%` = 0.75,
    `97.5\%` = 0.975, max = 1),
  names = TRUE
)
}
\arguments{
\item{x}{Numeric vector.}

\item{statfuns}{Function returning named summary statistics.}

\item{probs}{Named numeric vector of quantile probabilities.}

\item{names}{Logical; include names in quantile output.}
}
\value{
Named numeric vector of statistics and quantiles.
}
\description{
Computes summary statistics and quantiles for a numeric vector.
}
\details{
Quantile summary statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{MAE}
\alias{MAE}
\alias{MAE.data.frame}
\alias{MAE.matrix}
\alias{MAE.numeric}
\title{Mean Absolute Error}
\usage{
MAE(est, obs, ...)

\method{MAE}{data.frame}(est, obs, ...)

\method{MAE}{matrix}(est, obs, ...)

\method{MAE}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Passed to \code{mean}.}
}
\value{
Numeric MAE value.
}
\description{
Mean Absolute Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{just_CCC}
\alias{just_CCC}
\alias{just_CCC.data.frame}
\alias{just_CCC.matrix}
\alias{just_CCC.numeric}
\title{Lin's CCC (simplified)}
\usage{
just_CCC(y, x, ...)

\method{just_CCC}{data.frame}(y, x, ...)

\method{just_CCC}{matrix}(y, x, ...)

\method{just_CCC}{numeric}(y, x, ...)
}
\arguments{
\item{y}{Numeric vector, matrix, or data frame.}

\item{x}{Numeric vector, matrix, or data frame.}

\item{...}{Passed to \code{\link{LCCC}}.}
}
\value{
Numeric CCC value.
}
\description{
Lin's concordance correlation coefficient (simplified)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{quant_fun_groups}
\alias{quant_fun_groups}
\alias{quant_funs_groups}
\title{Quantile Summaries by Group}
\usage{
quant_fun_groups(x, catvar, pooled = TRUE, n = TRUE)

quant_funs_groups(x, catvar, pooled = TRUE, n = TRUE)
}
\arguments{
\item{x}{A list of numeric vectors.}

\item{catvar}{Factor or character vector of group labels.}

\item{pooled}{Logical; include a pooled row. Default \code{TRUE}.}

\item{n}{Logical; include count and missing count columns. Default \code{TRUE}.}
}
\value{
Data frame with one row per group.
}
\description{
Computes quantile summaries within each level of a categorical
  variable.
}
\details{
Quantile summaries by group
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{erMAX}
\alias{erMAX}
\alias{erMAX.data.frame}
\alias{erMAX.matrix}
\alias{erMAX.numeric}
\title{Maximum Absolute Relative Error}
\usage{
erMAX(est, obs, ...)

\method{erMAX}{data.frame}(est, obs, ...)

\method{erMAX}{matrix}(est, obs, ...)

\method{erMAX}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Currently unused.}
}
\value{
Numeric erMAX value.
}
\description{
Maximum Absolute Relative Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.xreg2}
\alias{summary.xreg2}
\title{Summarise an xreg2 Model}
\usage{
\method{summary}{xreg2}(object, ...)
}
\arguments{
\item{object}{A fitted \code{xreg2} object.}

\item{...}{Currently unused.}
}
\value{
\code{object} invisibly.
}
\description{
Prints a detailed summary of a fitted \code{xreg2} model.
  Currently equivalent to \code{\link{print.xreg2}}.
}
\details{
Summary method for xreg2 objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xreg2_fit.R
\name{xreg2_fit}
\alias{xreg2_fit}
\title{Fit xreg2 Model}
\usage{
xreg2_fit(
  controlList,
  dataList = NULL,
  start_values = numeric(),
  fixed_values = numeric(),
  latent_classes = 0,
  latent_class_parameters = character(),
  latent_id_colname = character(),
  return_type = "fit",
  print_sum = FALSE,
  method = "ucminf",
  hessian = TRUE,
  run_from = NULL,
  ...
)
}
\arguments{
\item{controlList}{An \code{xreg2ControlList} (produced by combining
\code{xreg2_control} objects with \code{c()}), an \code{xreg2_obj} (for
re-optimisation), or an \code{xreg2}/\code{xreg} fitted object (for
re-fitting).}

\item{dataList}{Named list of data frames, one per component. Names must
match those of \code{controlList}.}

\item{start_values}{Optional named numeric vector of starting values,
overriding those in \code{controlList}.}

\item{fixed_values}{Named numeric vector of parameter values held fixed.}

\item{latent_classes}{Integer; number of latent classes. \code{0} disables.}

\item{latent_class_parameters}{Character vector of parameter names that
vary by latent class.}

\item{latent_id_colname}{Character name of the grouping column for latent
classes.}

\item{return_type}{Character; \code{"fit"} (default), \code{"control"},
\code{"df"}, or \code{"predict"}.}

\item{print_sum}{Logical; print summary during optimisation. Default
\code{FALSE}.}

\item{method}{Character; optimisation method. Default \code{"ucminf"}.}

\item{hessian}{Logical; compute Hessian for standard errors. Default
\code{TRUE}.}

\item{run_from}{Internal; character string identifying the calling wrapper
(e.g. \code{"hyreg"}). Used to suppress certain warnings when called
from higher-level functions.}

\item{...}{Additional arguments passed to the optimiser.}
}
\value{
An object of class \code{"xreg2"} with fitted coefficients,
  standard errors, log-likelihoods, and counts.
}
\description{
Fits a flexible maximum-likelihood regression model supporting
  censored, interval, and hybrid data. Accepts a list of
  \code{\link{xreg2_control}} objects and a matching named list of
  data frames.
}
\details{
Fit a flexible maximum-likelihood regression model
}
\examples{
\donttest{
set.seed(1)
df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
ctrl <- xreg2_control(
  formulas     = list(y ~ INTERCEPT + x * SLOPE),
  start_values = c(INTERCEPT = 0, SLOPE = 0),
  name         = "main"
)
fit  <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))
coef(fit)
}

}
\seealso{
\code{\link{xreg2_control}}, \code{\link{xreg2_optim}},
  \code{\link{predict.xreg2}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xreg2_control.R
\name{c.xreg2Control}
\alias{c.xreg2Control}
\alias{c.xreg2ControlList}
\title{Combine xreg2 Control Objects}
\usage{
\method{c}{xreg2Control}(...)

\method{c}{xreg2ControlList}(...)
}
\arguments{
\item{...}{One or more \code{xreg2Control} objects. Named arguments are
used as component names; unnamed arguments use the object's \code{name}
field or a sequential integer.}
}
\value{
An object of class \code{c("xreg2ControlList", "list")}.
}
\description{
Combines one or more \code{xreg2Control} objects into an
  \code{xreg2ControlList} suitable for passing to \code{\link{xreg2_fit}}.
}
\details{
Combine xreg2Control objects into a list
}
\examples{
ctrl1 <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
                        name = "component1")
cl    <- c(ctrl1)

}
\seealso{
\code{\link{xreg2_control}}, \code{\link{xreg2_fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{ReMSE}
\alias{ReMSE}
\alias{ReMSE.data.frame}
\alias{ReMSE.matrix}
\alias{ReMSE.numeric}
\alias{ReMSE_conform}
\title{Relative Mean Squared Error}
\usage{
ReMSE(est, obs, trueobs, ...)

\method{ReMSE}{data.frame}(est, obs, trueobs, ...)

\method{ReMSE}{matrix}(est, obs, trueobs, ...)

\method{ReMSE}{numeric}(est, obs, trueobs, ...)

ReMSE_conform(dfys, dfobs, trueobs)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{trueobs}{Numeric; true values.}

\item{...}{Currently unused.}

\item{dfys}{Numeric matrix or data frame of estimates.}

\item{dfobs}{Numeric matrix or data frame of observations.}
}
\value{
Numeric ReMSE value.
}
\description{
Relative Mean Squared Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{comp_table}
\alias{comp_table}
\title{Comparison Table}
\usage{
comp_table(est, obs, ...)
}
\arguments{
\item{est}{Numeric vector or matrix of estimates.}

\item{obs}{Numeric vector of observations.}

\item{...}{Passed to metric functions.}
}
\value{
A numeric matrix with one row per metric.
}
\description{
Computes multiple error metrics and returns them as a matrix.
}
\details{
Comparison table of error metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{is.wholenumber}
\alias{is.wholenumber}
\title{Test for Whole Number}
\usage{
is.wholenumber(x, tol = .Machine$double.eps^0.5)
}
\arguments{
\item{x}{Numeric vector.}

\item{tol}{Numeric tolerance. Default \code{.Machine$double.eps^0.5}.}
}
\value{
Logical vector.
}
\description{
Tests whether numeric values are whole numbers within a
  numeric tolerance.
}
\details{
Test for whole number
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xreg2-package.R
\docType{package}
\name{xreg2-package}
\alias{xreg2}
\alias{xreg2-package}
\title{xreg2: Flexible Maximum Likelihood Regression with Gradient-Based Optimisation}
\description{
Fits flexible maximum likelihood regression models supporting
  censored, interval, and hybrid continuous/dichotomous data. Provides
  explicit analytic and numerical gradient computation, random intercept
  models via Gauss-Hermite quadrature, and multiple distribution families.
}
\author{
\strong{Maintainer}: Kim Rand \email{krand@krand.no}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{combsd}
\alias{combsd}
\title{Combined standard deviation of two groups}
\usage{
combsd(m1, m2, s1, s2, n1, n2, na.rm = FALSE)
}
\arguments{
\item{m1}{Mean of group 1.}

\item{m2}{Mean of group 2.}

\item{s1}{SD of group 1.}

\item{s2}{SD of group 2.}

\item{n1}{Size of group 1.}

\item{n2}{Size of group 2.}

\item{na.rm}{Logical.}
}
\value{
Combined SD.
}
\description{
Combined standard deviation of two groups
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{t_test2}
\alias{t_test2}
\title{Two-Sample t-Test from Summary Statistics}
\usage{
t_test2(m1, m2, s1, s2, n1, n2, m0 = 0, equal.variance = FALSE)
}
\arguments{
\item{m1}{Numeric; mean of sample 1.}

\item{m2}{Numeric; mean of sample 2.}

\item{s1}{Numeric; SD of sample 1.}

\item{s2}{Numeric; SD of sample 2.}

\item{n1}{Numeric; size of sample 1.}

\item{n2}{Numeric; size of sample 2.}

\item{m0}{Numeric; null difference. Default \code{0}.}

\item{equal.variance}{Logical; assume equal variances. Default \code{FALSE}.}
}
\value{
Named numeric vector with test statistics and confidence interval.
}
\description{
Two-sample t-test from summary statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{predict.xreg2}
\alias{predict.xreg2}
\title{Predict from an xreg2 Model}
\usage{
\method{predict}{xreg2}(object, newdata = NULL, return_vector = FALSE, ...)
}
\arguments{
\item{object}{A fitted \code{xreg2} object.}

\item{newdata}{A named list of data frames (one per model component), or a
single data frame (automatically wrapped and named after the first
component).}

\item{return_vector}{Logical; if \code{TRUE} and there is a single
component, return a numeric vector of \code{Xb} values instead of a
list. Default \code{FALSE}.}

\item{...}{Currently unused.}
}
\value{
A named list of data frames (one per component), each augmented
  with computed variables including \code{Xb}.
}
\description{
Generates predictions (linear predictors and intermediate
  computed variables) from a fitted \code{xreg2} model applied to new data.
}
\details{
Predict method for xreg2 objects
}
\examples{
\donttest{
set.seed(1)
df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
ctrl <- xreg2_control(
  formulas     = list(y ~ INTERCEPT + x * SLOPE),
  start_values = c(INTERCEPT = 0, SLOPE = 0),
  name         = "main"
)
fit  <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df))
newdf <- data.frame(x = 1:5)
preds <- predict(fit, newdata = list(main = newdf))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{ordE}
\alias{ordE}
\title{Order Statistic Expected Value}
\usage{
ordE(r, n, par = c(mean = 0, sd = 1), dist = "norm", log = FALSE)
}
\arguments{
\item{r}{Integer rank (or vector of ranks).}

\item{n}{Sample size.}

\item{par}{Named vector of distribution parameters.}

\item{dist}{Character distribution name.}

\item{log}{Currently unused.}
}
\value{
Numeric expected value.
}
\description{
Computes the expected value of the \eqn{r}-th order statistic
  in a random sample of \eqn{n} from a given distribution.
}
\details{
Expected value of an order statistic
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{MAPE}
\alias{MAPE}
\alias{MARE}
\alias{MARE.data.frame}
\alias{MARE.matrix}
\alias{MARE.numeric}
\title{Mean Absolute Relative Error (MARE / MAPE)}
\usage{
MAPE(est, obs, ...)

MARE(est, obs, ...)

\method{MARE}{data.frame}(est, obs, ...)

\method{MARE}{matrix}(est, obs, ...)

\method{MARE}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Currently unused.}
}
\value{
Numeric MARE value.
}
\description{
Mean Absolute Relative Error / Mean Absolute Percentage Error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{quant_funst}
\alias{quant_funst}
\title{Transposed Quantile Summaries (Matrix)}
\usage{
quant_funst(
  x,
  MARGIN = 1,
  statfuns = function(y) c(MEAN = mean(y, na.rm = TRUE), SE = sd(y, na.rm = TRUE)),
  probs = c(min = 0, `2.5\%` = 0.025, `25\%` = 0.25, median = 0.5, `75\%` = 0.75,
    `97.5\%` = 0.975, max = 1)
)
}
\arguments{
\item{x}{Matrix or data frame.}

\item{MARGIN}{Integer; \code{1} for rows, \code{2} for columns.}

\item{statfuns}{Function returning named summary statistics.}

\item{probs}{Named numeric vector of quantile probabilities.}
}
\value{
Transposed data frame of statistics.
}
\description{
Like \code{\link{quant_funs}} but returns transposed result.
}
\details{
Transposed quantile summaries over matrix margins
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{U95}
\alias{U95}
\alias{U95.data.frame}
\alias{U95.matrix}
\alias{U95.numeric}
\title{Uncertainty at 95 Percent}
\usage{
U95(est, obs, ...)

\method{U95}{data.frame}(est, obs, ...)

\method{U95}{matrix}(est, obs, ...)

\method{U95}{numeric}(est, obs, ...)
}
\arguments{
\item{est}{Numeric; estimates.}

\item{obs}{Numeric; observations.}

\item{...}{Passed to \code{var} and \code{sqrt}.}
}
\value{
Numeric U95 value.
}
\description{
Uncertainty at 95 Percent
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ll_functions.R
\name{xreg2_cont}
\alias{xreg2_cont}
\title{General Continuous Negative Log-Likelihood}
\usage{
xreg2_cont(
  par,
  dtalist,
  log.p = TRUE,
  aggregate.p = TRUE,
  print_pars = FALSE,
  dist = "normal"
)
}
\arguments{
\item{par}{Named numeric vector of current parameter values.}

\item{dtalist}{Internal data list produced by \code{xreg2_fit}.}

\item{log.p}{Logical; use log-scale probabilities. Default \code{TRUE}.}

\item{aggregate.p}{Logical; return scalar negloglik. Default \code{TRUE}.}

\item{print_pars}{Logical; print parameters and value. Default \code{FALSE}.}

\item{dist}{Character; distribution family. One of \code{"normal"},
\code{"lognormal"}, \code{"exp"}/\code{"exponential"}, \code{"beta"},
\code{"gamma"}, \code{"weibull"}. Default \code{"normal"}.}
}
\value{
Scalar negative log-likelihood (when \code{aggregate.p = TRUE}).
}
\description{
Computes the negative log-likelihood for a censored/interval
  regression model supporting multiple distributions (normal, lognormal,
  exponential, beta, gamma, weibull). Designed as a \code{p_fun} argument
  for \code{\link{xreg2_control}}.
}
\details{
General continuous log-likelihood for xreg2
}
\examples{
# ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#                        p_fun = xreg2_cont,
#                        start_values = c(INTERCEPT = 0, SLOPE = 0, LN_SIGMA = 0))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ll_functions.R
\name{xreg2_cont_r_lnorm}
\alias{xreg2_cont_r_lnorm}
\title{Continuous Log-Normal Negative Log-Likelihood with Random Intercept}
\usage{
xreg2_cont_r_lnorm(par, dtalist)
}
\arguments{
\item{par}{Named numeric vector of current parameter values. Must include
either \code{LN_OMEGA} or \code{OMEGA}.}

\item{dtalist}{Internal data list produced by \code{xreg2_fit}. Must
include an \code{internal_id} column.}
}
\value{
Scalar negative log-likelihood.
}
\description{
Computes the negative log-likelihood for a mixed-effects
  log-normal regression model with a random intercept, integrated via
  Gauss-Hermite quadrature. Designed as a \code{p_fun} argument for
  \code{\link{xreg2_control}}.
}
\details{
Continuous log-normal log-likelihood with random intercept
}
\examples{
# ctrl <- xreg2_control(formulas = list(y ~ INTERCEPT + x * SLOPE),
#                        p_fun = xreg2_cont_r_lnorm)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.xreg2}
\alias{print.xreg2}
\title{Print an xreg2 Model}
\usage{
\method{print}{xreg2}(x, ...)
}
\arguments{
\item{x}{A fitted \code{xreg2} object.}

\item{...}{Currently unused.}
}
\value{
\code{x} invisibly.
}
\description{
Prints a concise summary of a fitted \code{xreg2} model,
  including coefficients, standard errors, and minimised negative
  log-likelihoods.
}
\details{
Print method for xreg2 objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{combmean}
\alias{combmean}
\title{Combined mean of two groups}
\usage{
combmean(m1, m2, n1 = NULL, n2 = NULL, na.rm = FALSE)
}
\arguments{
\item{m1}{Mean of group 1 (scalar or vector).}

\item{m2}{Mean of group 2 (scalar or vector).}

\item{n1}{Size of group 1.}

\item{n2}{Size of group 2.}

\item{na.rm}{Logical.}
}
\value{
Combined mean.
}
\description{
Combined mean of two groups
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xreg2_control.R
\name{xreg2_control}
\alias{xreg2_control}
\title{Create xreg2 Control Object}
\usage{
xreg2_control(
  formulas,
  start_values = numeric(),
  fixed_values = numeric(),
  p_fun = xreg2_cont_normal,
  grad_fun = function(par, ...) numDeriv::grad(func = p_fun, x = par, method = "simple",
    ...),
  p_aggregation_fun = function(d_df) return(-d_df$p * d_df$internal_count),
  weights_var = NA,
  name = NA,
  censor_bounds = c(-Inf, Inf),
  lower = NA,
  upper = NA
)
}
\arguments{
\item{formulas}{A formula or list of formulas. The left-hand side names the
dependent (or intermediate) variable; the right-hand side is evaluated in
the context of parameter values and data columns.}

\item{start_values}{Named numeric vector of starting values. An existing
\code{xreg2} fitted object may be supplied; coefficients are extracted.}

\item{fixed_values}{Named numeric vector of parameter values held fixed
during optimisation.}

\item{p_fun}{Log-likelihood function. Default \code{xreg2_cont_normal}.
Must accept \code{(par, dtalist, ...)} and return a scalar.}

\item{grad_fun}{Gradient function. Default uses \code{numDeriv::grad} with
method \code{"simple"}.}

\item{p_aggregation_fun}{Aggregation function applied to the likelihood
data-frame. Rarely changed by users.}

\item{weights_var}{Character name of a column in the data used as
observation weights. \code{NA} for unweighted.}

\item{name}{Optional character label for this component.}

\item{censor_bounds}{Numeric vector of length 2 giving lower and upper
censoring bounds. Default \code{c(-Inf, Inf)} (no censoring).}

\item{lower}{Named numeric vector of lower parameter bounds (for
bounded optimisers).}

\item{upper}{Named numeric vector of upper parameter bounds.}
}
\value{
An object of class \code{c("xreg2Control", "list")}.
}
\description{
Creates a control object specifying the model formula(s),
  likelihood function, gradient function, and starting values for one
  component of a joint \code{xreg2} model.
}
\details{
Create an xreg2 control object
}
\examples{
ctrl <- xreg2_control(
  formulas     = list(y ~ INTERCEPT + x * SLOPE),
  start_values = c(INTERCEPT = 0, SLOPE = 0),
  p_fun        = xreg2_cont_normal,
  name         = "continuous"
)

}
\seealso{
\code{\link{xreg2_fit}}, \code{\link{c.xreg2Control}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{fisher_icc}
\alias{fisher_icc}
\alias{fisher_icc.data.frame}
\alias{fisher_icc.matrix}
\alias{fisher_icc.numeric}
\title{Fisher's ICC}
\usage{
fisher_icc(y, x, ...)

\method{fisher_icc}{data.frame}(y, x, ...)

\method{fisher_icc}{matrix}(y, x, ...)

\method{fisher_icc}{numeric}(y, x, ...)
}
\arguments{
\item{y}{Numeric vector, matrix, or data frame.}

\item{x}{Numeric vector, matrix, or data frame.}

\item{...}{Currently unused.}
}
\value{
Numeric ICC value.
}
\description{
Computes Fisher's intraclass correlation coefficient between
  two numeric vectors or matrices.
}
\details{
Fisher's intraclass correlation coefficient
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{getMean}
\alias{getMean}
\title{Mean of a distribution}
\usage{
getMean(qfun, pars)
}
\arguments{
\item{qfun}{Quantile function.}

\item{pars}{Named list of distribution parameters.}
}
\value{
Numeric mean.
}
\description{
Mean of a distribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat_methods.R
\name{ordD}
\alias{ordD}
\title{Order Statistic Density}
\usage{
ordD(x, r, n, par = c(mean = 0, sd = 1), dist = "norm", log = FALSE)
}
\arguments{
\item{x}{Numeric vector of values.}

\item{r}{Integer rank.}

\item{n}{Sample size.}

\item{par}{Named vector of distribution parameters.}

\item{dist}{Character distribution name.}

\item{log}{Logical; return log-density. Default \code{FALSE}.}
}
\value{
Numeric density (or log-density).
}
\description{
Computes the density of the \eqn{r}-th order statistic in a
  random sample of \eqn{n} from a given distribution.
}
\details{
Density for an order statistic
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xreg2_fit.R
\name{xreg2_optim}
\alias{xreg2_optim}
\title{Optimise an xreg2 Model}
\usage{
xreg2_optim(
  xreg2obj,
  method = NULL,
  hessian = TRUE,
  fixed_values = NULL,
  optim_control = NULL,
  ...
)
}
\arguments{
\item{xreg2obj}{An \code{xreg2_obj} list produced by
\code{\link{xreg2_fit}} with \code{return_type = "control"}.}

\item{method}{Character; optimisation method. \code{NULL} defaults to
\code{"ucminf"}.}

\item{hessian}{Logical; compute Hessian. Default \code{TRUE}.}

\item{fixed_values}{Optional named numeric vector of additional fixed values.}

\item{optim_control}{Optional named list of control parameters that override
the method-specific defaults set by \code{\link{xreg2_fit}}.  For example,
\code{optim_control = list(maxit = 500)} tightens the iteration budget.
Passed directly to \code{\link[ucminf]{ucminf}}, \code{\link[stats]{optim}},
or \code{optimx::optimr} depending on \code{method}.}

\item{...}{Additional arguments passed to the optimiser objective function.}
}
\value{
An object of class \code{"xreg2"}.
}
\description{
Low-level optimisation routine called by \code{\link{xreg2_fit}}
  and \code{\link{xreg2_fit_s}}. Can also be called directly on an
  \code{xreg2_obj} to re-optimise with different settings.

  Optimizer routing:
  \itemize{
    \item \code{"ucminf"} (default) — uses \pkg{ucminf} with central-difference
      gradients and high-precision tolerances.
    \item \code{"L-BFGS-B"}, \code{"BFGS"}, \code{"CG"}, \code{"Nelder-Mead"},
      \code{"SANN"}, \code{"Brent"} — uses \code{stats::optim()}.
      Box constraints (\code{lower}/\code{upper} set in
      \code{\link{xreg2_control}}) are enforced for \code{"L-BFGS-B"}.
    \item Any other method string — requires the \pkg{optimx} package (listed
      in \code{Suggests}) and dispatches via \code{optimx::optimr()}.
  }

  Method-specific control defaults (high-precision) are set automatically and
  can be overridden by passing \code{optim_control = list(...)} to
  \code{\link{xreg2_fit}}.
}
\details{
Optimise an xreg2 model object
}
\examples{
\donttest{
set.seed(1)
df   <- data.frame(y = rnorm(200, 2, 0.5), x = rnorm(200))
ctrl <- xreg2_control(
  formulas     = list(y ~ INTERCEPT + x * SLOPE),
  start_values = c(INTERCEPT = 0, SLOPE = 0),
  name         = "main"
)
obj          <- xreg2_fit(controlList = c(ctrl), dataList = list(main = df),
                          return_type = "control")
fit          <- xreg2_optim(obj, method = "ucminf")
fit_lbfgsb   <- xreg2_optim(obj, method = "L-BFGS-B")
}

}
\seealso{
\code{\link{xreg2_fit}}
}
